# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import copy
import io
import itertools
import operator
import os
import re
import shlex

import numpy

from .. import config
from . import elements
from . import spacegrouplattice as sgl
from . import wyckpos

re_data = re.compile(r"^data_", re.IGNORECASE)
re_loop = re.compile(r"^loop_", re.IGNORECASE)
re_symop = re.compile(r"^\s*("
                      "_space_group_symop_operation_xyz|"
                      "_symmetry_equiv_pos_as_xyz)", re.IGNORECASE)
re_name = re.compile(r"^\s*_chemical_formula_sum", re.IGNORECASE)
re_atom = re.compile(r"^\s*_atom_site_label\s*$", re.IGNORECASE)
re_atomtyp = re.compile(r"^\s*_atom_site_type_symbol\s*$", re.IGNORECASE)
re_atomx = re.compile(r"^\s*_atom_site_fract_x", re.IGNORECASE)
re_atomy = re.compile(r"^\s*_atom_site_fract_y", re.IGNORECASE)
re_atomz = re.compile(r"^\s*_atom_site_fract_z", re.IGNORECASE)
re_uiso = re.compile(r"^\s*_atom_site_U_iso_or_equiv", re.IGNORECASE)
re_biso = re.compile(r"^\s*_atom_site_B_iso_or_equiv", re.IGNORECASE)
re_atomocc = re.compile(r"^\s*_atom_site_occupancy", re.IGNORECASE)
re_labelline = re.compile(r"^\s*_")
re_emptyline = re.compile(r"^\s*$")
re_quote = re.compile(r"'")
re_spacegroupnr = re.compile(r"^\s*(_space_group_IT_number|"
                             "_symmetry_Int_Tables_number)", re.IGNORECASE)
re_spacegroupname = re.compile(r"^\s*(_symmetry_space_group_name_H-M|"
                               "_space_group_name_H-M_alt)",
                               re.IGNORECASE)
re_spacegroupsetting = re.compile(r"^\s*_symmetry_cell_setting", re.IGNORECASE)
re_cell_a = re.compile(r"^\s*_cell_length_a", re.IGNORECASE)
re_cell_b = re.compile(r"^\s*_cell_length_b", re.IGNORECASE)
re_cell_c = re.compile(r"^\s*_cell_length_c", re.IGNORECASE)
re_cell_alpha = re.compile(r"^\s*_cell_angle_alpha", re.IGNORECASE)
re_cell_beta = re.compile(r"^\s*_cell_angle_beta", re.IGNORECASE)
re_cell_gamma = re.compile(r"^\s*_cell_angle_gamma", re.IGNORECASE)
re_comment = re.compile(r"^\s*#")


class CIFFile(object):
    """
    class for parsing CIF (Crystallographic Information File) files. The class
    aims to provide an additional way of creating material classes instead of
    manual entering of the information the lattice constants and unit cell
    structure are parsed from the CIF file.

    If multiple datasets are present in the CIF file this class will attempt to
    parse all of them into the the data dictionary. By default all methods
    access the first data set found in the file.
    """
    def __init__(self, filestr, digits=4):
        """
        initialization of the CIFFile class

        Parameters
        ----------
        filestr :  str, bytes
            CIF filename or string representation of the CIF file
        digits :    int, optional
            number of digits to check if position is unique
        """
        self.digits = digits

        if os.path.isfile(filestr):
            self.filename = filestr
            try:
                self.fid = open(self.filename, "rb")
            except OSError:
                raise IOError("cannot open CIF file %s" % self.filename)
        else:
            if filestr.count('\n') == 0:
                print('XU.materials.CIFFile: "filestr" contains only one line '
                      'but a file with that name does not exist! Continuing '
                      'with the assumption this one line string is the '
                      'content of a CIF file!')
            self.filename = '__from_str__'
            if isinstance(filestr, bytes):
                self.fid = io.BytesIO(filestr)
            else:
                self.fid = io.BytesIO(bytes(filestr.encode('ascii')))

        if config.VERBOSITY >= config.INFO_ALL:
            print('XU.materials: parsing CIF file %s' % self.filename)
        self._default_dataset = None
        self.data = {}
        self.Parse()

    def __del__(self):
        """
        class destructor which closes open files
        """
        if self.fid is not None:
            self.fid.close()

    def Parse(self):
        """
        function to parse a CIF file. The function reads all the included data
        sets and adds them to the data dictionary.

        """
        fidpos = self.fid.tell()
        while True:
            line = self.fid.readline()
            if not line:
                break
            fidpos = self.fid.tell()
            line = line.decode('ascii', 'ignore')
            m = re_data.match(line)
            if m:
                self.fid.seek(fidpos)
                name = line[m.end():].strip()
                self.data[name] = CIFDataset(self.fid, name, self.digits)
                if self.data[name].has_atoms and not self._default_dataset:
                    self._default_dataset = name

    def SGLattice(self, dataset=None, use_p1=False):
        """
        create a SGLattice object with the structure from the CIF dataset

        Parameters
        ----------
        dataset :   str, optional
            name of the dataset to use. if None the default one will be used.
        use_p1 :    bool, optional
            force the use of P1 symmetry, default False
        """
        if not dataset:
            dataset = self._default_dataset
        return self.data[dataset].SGLattice(use_p1=use_p1)

    def __str__(self):
        """
        returns a string with positions and names of the atoms for all datasets
        """
        ostr = ""
        ostr += "CIF-File: %s\n" % self.filename
        for ds in self.data:
            ostr += "\nDataset: %s" % ds
            if ds == self._default_dataset:
                ostr += " (default)"
            ostr += "\n"
            ostr += str(self.data[ds])
        return ostr


class CIFDataset(object):
    """
    class for parsing CIF (Crystallographic Information File) files. The class
    aims to provide an additional way of creating material classes instead of
    manual entering of the information the lattice constants and unit cell
    structure are parsed from the CIF file
    """

    def __init__(self, fid, name, digits):
        """
        initialization of the CIFDataset class. This class parses one data
        block.

        Parameters
        ----------
        fid :       filehandle
            file handle set to the beginning of the data block to be parsed
        name :      str
            identifier string of the dataset
        digits :    int
            number of digits to check if position is unique
        """
        self.name = name
        self.digits = digits
        self.has_atoms = False

        if config.VERBOSITY >= config.INFO_ALL:
            print('XU.materials: parsing cif dataset %s' % self.name)
        self.Parse(fid)
        self.SymStruct()

    def Parse(self, fid):
        """
        function to parse a CIF data set. The function reads the
        space group symmetry operations and the basic atom positions
        as well as the lattice constants and unit cell angles
        """

        self.symops = []
        self.atoms = []
        self.lattice_const = numpy.zeros(3, dtype=numpy.double)
        self.lattice_angles = numpy.zeros(3, dtype=numpy.double)

        loop_start = False
        symop_loop = False
        atom_loop = False

        def get_element(cifstring):
            el = re.sub(r"['\"]", r"", cifstring)
            if '+' in el or '-' in el:
                # add oxidation number if not present
                for sign in ('+', '-'):
                    if sign in el:
                        if not el[el.index(sign)-1].isdigit():
                            signidx = el.index(sign)
                            el = el[:signidx] + '1' + el[signidx:]
                # replace special characters
                for r, o in zip(('dot', 'p', 'm'), ('.', '+', '-')):
                    el = el.replace(o, r)
            else:
                el = re.sub(r"([0-9])", r"", el)
            el = re.sub(r"\(\w*\)", r"", el)
            try:
                element = getattr(elements, el)
            except AttributeError:  # el not found, typ. due to oxidation state
                f = re.search('[0-9]', el)
                if not f and el == '?':
                    element = elements.Dummy
                elif f is None:
                    raise ValueError("XU.materials: element ('%s') could not"
                                     " be identified as chemical element. Only"
                                     " abbreviations of element names are "
                                     "supported." % (cifstring))
                else:
                    elname = el[:f.start()]
                    if hasattr(elements, elname):
                        # here one might want to find a closer alternative than
                        # the neutral atom, but the effect this has should be
                        # minimal, currently simply the neutral atom is used
                        if config.VERBOSITY >= config.INFO_LOW:
                            print('XU.materials: element %s used instead of %s'
                                  % (elname, cifstring))
                        element = getattr(elements, elname)
                    else:
                        raise ValueError("XU.materials: element ('%s') could "
                                         "not be found" % (cifstring))
            return element

        def floatconv(string):
            """
            helper function to convert string with possible error
            given in brackets to float
            """
            try:
                f = float(re.sub(r"\(.+\)", r"", string))
            except ValueError:
                f = numpy.nan
            return f

        def intconv(string):
            """
            helper function to convert string to integer
            """
            try:
                i = int(string)
            except ValueError:
                i = None
            return i

        fidpos = fid.tell()
        for line in fid.readlines():
            linelen = len(line)
            line = line.decode('ascii', 'ignore')
            if config.VERBOSITY >= config.DEBUG:
                print(line)
                print(fid.tell(), fidpos)

            if re_data.match(line):
                fid.seek(fidpos)
                break
            fidpos += linelen

            # ignore comment lines
            if re_comment.match(line):
                continue

            if re_loop.match(line):  # start of loop
                if config.VERBOSITY >= config.DEBUG:
                    print('XU.materials: loop start found')
                loop_start = True
                loop_labels = []
                symop_loop = False
                atom_loop = False
                ax_idx = None
                ay_idx = None
                az_idx = None
                uiso_idx = None
                biso_idx = None
                occ_idx = None
            elif re_labelline.match(line):
                if re_cell_a.match(line):
                    self.lattice_const[0] = floatconv(line.split()[1])
                elif re_cell_b.match(line):
                    self.lattice_const[1] = floatconv(line.split()[1])
                elif re_cell_c.match(line):
                    self.lattice_const[2] = floatconv(line.split()[1])
                elif re_cell_alpha.match(line):
                    self.lattice_angles[0] = floatconv(line.split()[1])
                elif re_cell_beta.match(line):
                    self.lattice_angles[1] = floatconv(line.split()[1])
                elif re_cell_gamma.match(line):
                    self.lattice_angles[2] = floatconv(line.split()[1])
                elif re_spacegroupnr.match(line):
                    i = intconv(line.split()[1])
                    if i:
                        self.sgrp_nr = i
                elif re_spacegroupname.match(line):
                    self.sgrp_name = ''.join(line.split()[1:]).strip("'")
                elif re_spacegroupsetting.match(line):
                    i = intconv(line.split()[1])
                    if i:
                        self.sgrp_setting = i
                elif re_name.match(line):
                    try:
                        self.name = shlex.split(line)[1]
                    except IndexError:
                        self.name = None
                if loop_start:
                    loop_labels.append(line.strip())
                    if re_symop.match(line):  # start of symmetry op. loop
                        if config.VERBOSITY >= config.DEBUG:
                            print('XU.materials: symop-loop identified')
                        symop_loop = True
                        symop_idx = len(loop_labels) - 1
                    elif re_atom.match(line) or re_atomtyp.match(line):
                        # start of atom position loop
                        if config.VERBOSITY >= config.DEBUG:
                            print('XU.materials: atom position-loop found')
                        atom_loop = True
                        if re_atomtyp.match(line):
                            alab_idx = len(loop_labels) - 1
                        elif not list(filter(re_atomtyp.match, loop_labels)):
                            # ensure precedence of atom_site_type_symbol
                            alab_idx = len(loop_labels) - 1
                    elif re_atomx.match(line):
                        ax_idx = len(loop_labels) - 1
                        if config.VERBOSITY >= config.DEBUG:
                            print('XU.materials: atom position x: col%d'
                                  % ax_idx)
                    elif re_atomy.match(line):
                        ay_idx = len(loop_labels) - 1
                    elif re_atomz.match(line):
                        az_idx = len(loop_labels) - 1
                    elif re_uiso.match(line):
                        uiso_idx = len(loop_labels) - 1
                    elif re_biso.match(line):
                        biso_idx = len(loop_labels) - 1
                    elif re_atomocc.match(line):
                        occ_idx = len(loop_labels) - 1

            elif re_emptyline.match(line):
                loop_start = False
                symop_loop = False
                atom_loop = False
                continue
            elif symop_loop:  # symmetry operation entry
                loop_start = False
                entry = shlex.split(line)[symop_idx]
                if re_quote.match(entry):
                    opstr = entry
                else:
                    opstr = "'" + entry + "'"
                opstr = re.sub(r"^'", r"(", opstr)
                opstr = re.sub(r"'$", r")", opstr)
                self.symops.append(opstr)
            elif atom_loop:  # atom label and position
                loop_start = False
                asplit = line.split()
                try:
                    atom = get_element(asplit[alab_idx])
                    apos = (floatconv(asplit[ax_idx]),
                            floatconv(asplit[ay_idx]),
                            floatconv(asplit[az_idx]))
                    occ = floatconv(asplit[occ_idx]) if occ_idx else 1
                    if numpy.isnan(occ):
                        occ = 1
                    uiso = floatconv(asplit[uiso_idx]) if uiso_idx else 0
                    biso = floatconv(asplit[biso_idx]) if biso_idx else 0
                    if numpy.isnan(uiso):
                        uiso = 0
                    if numpy.isnan(biso):
                        biso = 0
                    if biso == 0:
                        biso = 8 * numpy.pi**2 * uiso
                    self.atoms.append((atom, apos, occ, biso))
                except IndexError:
                    if config.VERBOSITY >= config.INFO_LOW:
                        print('XU.materials: could not parse atom line: "%s"'
                              % line.strip())
        if self.atoms:
            self.has_atoms = True

    def SymStruct(self):
        """
        function to obtain the list of different atom positions in the unit
        cell for the different types of atoms and determine the space group
        number and origin choice if available. The data are obtained from the
        data parsed from the CIF file.
        """

        def rem_white(string):
            return string.replace(' ', '')

        if hasattr(self, 'sgrp_name'):
            # determine spacegroup
            cifsgn = rem_white(self.sgrp_name).split(':')[0]
            for nr, name in sgl.sgrp_name.items():
                if cifsgn == rem_white(name):
                    self.sgrp_nr = int(nr)
            if not hasattr(self, 'sgrp_nr'):
                # try ignoring the minuses
                for nr, name in sgl.sgrp_name.items():
                    if cifsgn == rem_white(name.replace('-', '')):
                        self.sgrp_nr = int(nr)
            if len(self.sgrp_name.split(':')) > 1:
                self.sgrp_suf = ':' + self.sgrp_name.split(':')[1]
            elif hasattr(self, 'sgrp_setting'):
                self.sgrp_suf = ':%d' % self.sgrp_setting

        if not hasattr(self, 'sgrp_suf'):
            if hasattr(self, 'sgrp_nr'):
                self.sgrp_suf = sgl.get_possible_sgrp_suf(self.sgrp_nr)
            else:
                self.sgrp_suf = ''
        if isinstance(self.sgrp_suf, str):
            suffixes = [self.sgrp_suf, ]
        else:
            suffixes = copy.copy(self.sgrp_suf)
        for sgrp_suf in suffixes:
            self.sgrp_suf = sgrp_suf
            if hasattr(self, 'sgrp_nr'):
                self.sgrp = str(self.sgrp_nr) + self.sgrp_suf
                allwyckp = wyckpos.wp[self.sgrp]
                if config.VERBOSITY >= config.INFO_ALL:
                    print('XU.materials: trying space group %s' % self.sgrp)

            # determine all unique positions for definition of a P1 space group
            symops = self.symops
            if not symops and hasattr(self, 'sgrp') and self.atoms:
                label = sorted(allwyckp, key=lambda s: int(s[:-1]))
                symops = allwyckp[label[-1]][1]
                if config.VERBOSITY >= config.INFO_ALL:
                    print('XU.materials: no symmetry operations in CIF-Dataset'
                          '; using built in general positions.')
            self.unique_positions = []
            for el, (x, y, z), occ, biso in self.atoms:
                unique_pos = set()
                for symop in symops:
                    pos = eval(symop, {'x': x, 'y': y, 'z': z})
                    pos = numpy.asarray(pos)
                    # check that position is within unit cell
                    pos = pos - numpy.round(pos, self.digits) // 1
                    unique_pos.add(tuple(numpy.round(pos, self.digits)))
                self.unique_positions.append((el, unique_pos, occ, biso))

            # determine Wyckoff positions and free parameters of unit cell
            if hasattr(self, 'sgrp'):
                self.wp = []
                self.occ = []
                self.elements = []
                self.biso = []
                keys = list(allwyckp)
                wpn = [int(re.sub(r"([a-zA-Z])", r"", k)) for k in keys]
                for i, (el, (x, y, z), occ, biso) in enumerate(self.atoms):
                    # candidate positions from number of unique atoms
                    natoms = len(self.unique_positions[i][1])
                    wpcand = []
                    for j, n in enumerate(wpn):
                        if n == natoms:
                            wpcand.append((keys[j], allwyckp[keys[j]]))
                    for j, (k, wp) in enumerate(
                            sorted(wpcand, key=operator.itemgetter(1))):
                        parint, poslist, reflcond = wp
                        for positem in poslist:
                            foundwp, xyz = sgl.testwp(parint, positem,
                                                      (x, y, z), self.digits)
                            if foundwp:
                                if xyz is None:
                                    self.wp.append(k)
                                else:
                                    self.wp.append((k, xyz))
                                self.elements.append(el)
                                self.occ.append(occ)
                                self.biso.append(biso)
                                break
                        if foundwp:
                            break
                if config.VERBOSITY >= config.INFO_ALL:
                    print('XU.materials: %d of %d Wyckoff positions identified'
                          % (len(self.wp), len(self.atoms)))
                    if len(self.wp) < len(self.atoms):
                        print('XU.materials: space group %s seems not to fit'
                              % self.sgrp)

                # free unit cell parameters
                self.crystal_system, nargs = sgl.sgrp_sym[self.sgrp_nr]
                self.crystal_system += self.sgrp_suf
                self.uc_params = []
                p2i = {'a': 0, 'b': 1, 'c': 2,
                       'alpha': 0, 'beta': 1, 'gamma': 2}
                for pname in sgl.sgrp_params[self.crystal_system][0]:
                    if pname in ('a', 'b', 'c'):
                        self.uc_params.append(self.lattice_const[p2i[pname]])
                    if pname in ('alpha', 'beta', 'gamma'):
                        self.uc_params.append(self.lattice_angles[p2i[pname]])

                if len(self.wp) == len(self.atoms):
                    if config.VERBOSITY >= config.INFO_ALL:
                        print('XU.materials: identified space group as %s'
                              % self.sgrp)
                    break

    def SGLattice(self, use_p1=False):
        """
        create a SGLattice object with the structure from the CIF file
        """
        if not use_p1:
            if hasattr(self, 'sgrp'):
                if len(self.wp) == len(self.atoms):
                    return sgl.SGLattice(self.sgrp, *self.uc_params,
                                         atoms=self.elements, pos=self.wp,
                                         occ=self.occ, b=self.biso)
                else:
                    if config.VERBOSITY >= config.INFO_LOW:
                        print('XU.materials: Wyckoff positions missing, '
                              'using P1')
            else:
                if config.VERBOSITY >= config.INFO_LOW:
                    print('XU.materials: space-group detection failed, '
                          'using P1')

        atomdict = {'atoms': [], 'pos': [], 'occ': [], 'b': []}
        for element, positions, o, b in self.unique_positions:
            for p in positions:
                atomdict['atoms'].append(element)
                atomdict['pos'].append(('1a', p))
                atomdict['occ'].append(o)
                atomdict['b'].append(b)

        return sgl.SGLattice(1, *itertools.chain(self.lattice_const,
                                                 self.lattice_angles),
                             **atomdict)

    def __str__(self):
        """
        returns a string with positions and names of the atoms
        """
        ostr = ""
        ostr += "unit cell structure:"
        if hasattr(self, 'sgrp'):
            ostr += " %s %s %s\n" % (self.sgrp, self.crystal_system,
                                     getattr(self, 'sgrp_name', ''))
        else:
            ostr += "\n"
        ostr += "a: %8.4f b: %8.4f c: %8.4f\n" % tuple(self.lattice_const)
        ostr += "alpha: %6.2f beta: %6.2f gamma: %6.2f\n" % tuple(
            self.lattice_angles)
        if self.unique_positions:
            ostr += "Unique atom positions in unit cell\n"
        for atom in self.unique_positions:
            ostr += atom[0].name + " (%d): \n" % atom[0].num
            for pos in atom[1]:
                ostr += str(numpy.round(pos, self.digits)) + "\n"
        return ostr


def cifexport(filename, mat):
    """
    function to export a Crystal instance to CIF file. This in particular
    includes the atomic coordinates, however, ignores for example the elastic
    parameters.
    """

    def unique_label(basename, names):
        num = 1
        name = '{name}{num:d}'.format(name=basename, num=num)
        while name in names:
            num += 1
            name = '{name}{num:d}'.format(name=basename, num=num)
        return name

    general = """data_global
_chemical_formula_sum '{chemsum}'
_cell_length_a {a:.5f}
_cell_length_b {b:.5f}
_cell_length_c {c:.5f}
_cell_angle_alpha {alpha:.4f}
_cell_angle_beta {beta:.4f}
_cell_angle_gamma {gamma:.4f}
_cell_volume {vol:.3f}
_space_group_crystal_system {csystem}
_space_group_IT_number {sgrpnr}
_space_group_name_H-M_alt '{hmsymb}'
"""

    csystem = mat.lattice.crystal_system
    if len(mat.lattice.space_group_suf) > 0:
        csystem = csystem[:-len(mat.lattice.space_group_suf)]

    ctxt = general.format(chemsum=mat.chemical_composition(with_spaces=True),
                          a=mat.a, b=mat.b, c=mat.c,
                          alpha=mat.alpha, beta=mat.beta, gamma=mat.gamma,
                          vol=mat.lattice.UnitCellVolume(),
                          csystem=csystem,
                          sgrpnr=mat.lattice.space_group_nr,
                          hmsymb=mat.lattice.name)

    sgrpsuf = mat.lattice.space_group_suf[1:]
    if sgrpsuf:
        ctxt += '_symmetry_cell_setting {suf}\n'.format(suf=sgrpsuf)

    symloop = """
loop_
_space_group_symop_operation_xyz
"""

    for symop in mat.lattice.symops:
        symloop += "'" + symop.xyz() + "'\n"

    atomloop = """
loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_symmetry_multiplicity
_atom_site_Wyckoff_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_occupancy
_atom_site_B_iso_or_equiv
"""
    nidx = 0
    allatoms = list(mat.lattice.base())
    names = []
    for at, pos, occ, b in mat.lattice._wbase:
        wm, wl, _ = re.split('([a-z])', pos[0])
        nsite = int(wm)
        x, y, z = allatoms[nidx][1]
        names.append(unique_label(at.name, names))
        atomloop += '%s %s %d %c %.5f %.5f %.5f %.4f %.4f\n' % (
            names[-1], at.name, nsite, wl, x, y, z, occ, b)
        nidx += nsite

    with open(filename, 'w') as f:
        f.write(ctxt)
        f.write(symloop)
        f.write(atomloop)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import time

import numpy

from .. import config
from ..exception import InputError
# relative imports from xrayutilities
from .helper import xu_open


class ImageReader(object):

    """
    parse CCD frames in the form of tiffs or binary data (``*.bin``)
    to numpy arrays. ignore the header since it seems to contain
    no useful data

    The routine was tested so far with

    1. RoperScientific files with 4096x4096 pixels created at Hasylab Hamburg,
       which save an 16bit integer per point.
    2. Perkin Elmer images created at Hasylab Hamburg with 2048x2048 pixels.
    """

    def __init__(self, nop1, nop2, hdrlen=0, flatfield=None, darkfield=None,
                 dtype=numpy.int16, byte_swap=False):
        """
        initialize the ImageReader reader, which includes setting the dimension
        of the images as well as defining the data used for flat- and darkfield
        correction!

        Parameters
        ----------
        nop1, nop2 :    int
            number of pixels in the first and second dimension of the image
        hdrlen :        int, optional
            length of the file header which should be ignored
        flatfield :     str or ndarray, optional
            filename or data for flatfield correction. supported file types
            include (.bin/.tif (also compressed .xz or .gz) and .npy files).
            otherwise a 2D numpy array should be given
        darkfield :     str or ndarray, optional
            filename or data for darkfield correction. same types as for flat
            field are supported.
        dtype :         numpy.dtype, optional
            datatype of the stored values (default: numpy.int16)
        byte_swap :     bool, optional
            flag which determines bytes are swapped after reading
        """

        # save number of pixels per image
        self.nop1 = nop1
        self.nop2 = nop2
        self.dtype = dtype
        self.hdrlen = hdrlen
        self.byteswap = byte_swap

        # read flatfield
        if flatfield:
            if isinstance(flatfield, str):
                if os.path.splitext(flatfield)[1] == '.npy':
                    self.flatfield = numpy.load(flatfield)
                elif os.path.splitext(flatfield)[1] in \
                        ['.gz', '.xz', '.bin', '.tif']:
                    # read without flatc and darkc
                    self.flatfield = self.readImage(flatfield)
                else:
                    raise InputError("Error: unknown filename for "
                                     "flatfield correction!")
            elif isinstance(flatfield, numpy.ndarray):
                self.flatfield = flatfield
            else:
                raise InputError("Error: unsupported type for "
                                 "flatfield correction!")

        # read darkfield
        if darkfield:
            if isinstance(darkfield, str):
                if os.path.splitext(darkfield)[1] == '.npy':
                    self.darkfield = numpy.load(darkfield)
                elif os.path.splitext(darkfield)[1] in \
                        ['.gz', '.xz', '.bin', '.tif']:
                    # read without flatc and darkc
                    self.darkfield = self.readImage(darkfield)
                else:
                    raise InputError("Error: unknown filename for "
                                     "darkfield correction!")
            elif isinstance(darkfield, numpy.ndarray):
                self.darkfield = darkfield
            else:
                raise InputError("Error: unsupported type for "
                                 "darkfield correction!")

        if flatfield:
            self.flatc = True
            if config.VERBOSITY >= config.INFO_ALL:
                print("XU.io.ImageReader: flatfield correction enabled")
        else:
            self.flatc = False
        if darkfield:
            self.darkc = True
            if config.VERBOSITY >= config.INFO_ALL:
                print("XU.io.ImageReader: darkfield correction enabled")
        else:
            self.darkc = False

    def readImage(self, filename, path=None):
        """
        read image file
        and correct for dark- and flatfield in case the necessary data are
        available.

        returned data = ((image data)-(darkfield))/flatfield*average(flatfield)

        Parameters
        ----------
        filename :  str
            filename of the image to be read. so far only single filenames are
            supported. The data might be compressed.  supported extensions:
            .tif, .bin and .bin.xz
        path :      str, optional
            path of the data files
        """
        if path:
            full_filename = os.path.join(path, filename)
        else:
            full_filename = filename

        if config.VERBOSITY >= config.INFO_ALL:
            print("XU.io.ImageReader.readImage: file %s" % (full_filename))
            t1 = time.time()

        with xu_open(full_filename) as fh:
            # jump over header
            fh.seek(self.hdrlen)
            # read image
            rlen = numpy.dtype(self.dtype).itemsize * self.nop1 * self.nop2
            img = numpy.frombuffer(fh.read(rlen), dtype=self.dtype)
            if self.byteswap:
                img = img.byteswap()
            img.shape = (self.nop1, self.nop2)  # reshape the data
            # darkfield correction
            if self.darkc:
                img = (img - self.darkfield).astype(numpy.float32)
            # flatfield correction
            if self.flatc:
                img = img.astype(numpy.float32) / self.flatfield

        if config.VERBOSITY >= config.INFO_ALL:
            t2 = time.time()
            print("XU.io.ImageReader.readImage: parsing time %8.3f"
                  % (t2 - t1))

        return img


dlen = {'char': 1,
        'byte': 1,
        'word': 2,
        'dword': 4,
        'rational': 8,  # not implemented correctly
        'float': 4,
        'double': 8}

dtypes = {1: 'byte',
          2: 'char',
          3: 'word',
          4: 'dword',
          5: 'rational',  # not implemented correctly
          6: 'byte',
          7: 'byte',
          8: 'word',
          9: 'dword',
          10: 'rational',  # not implemented correctly
          11: 'float',
          12: 'double'}

nptyp = {1: numpy.byte,
         2: numpy.char,
         3: numpy.uint16,
         4: numpy.uint32,
         5: numpy.uint32,
         6: numpy.int8,
         7: numpy.byte,
         8: numpy.int16,
         9: numpy.int32,
         10: numpy.int32,
         11: numpy.float32,
         12: numpy.float64}

tiffdtype = {1: {8:  numpy.uint8, 16: numpy.uint16, 32: numpy.uint32},
             2: {8:  numpy.int8, 16: numpy.int16, 32: numpy.int32},
             3: {16: numpy.float16, 32: numpy.float32}}

tifftags = {256: 'ImageWidth',  # width
            257: 'ImageLength',  # height
            258: 'BitsPerSample',
            259: 'Compression',
            262: 'PhotometricInterpretation',
            270: 'ImageDescription',
            272: 'Model',
            273: 'StripOffsets',
            278: 'RowsPerStrip',
            279: 'StripByteCount',
            282: 'XResolution',
            283: 'YResolution',
            305: 'Software',
            306: 'DateTime',
            315: 'Artist',
            339: 'SampleFormat',
            36864: 'ExifVersion'}


class TIFFRead(ImageReader):
    """
    class to Parse a TIFF file including extraction of information from the
    file header in order to determine the image size and data type

    The data stored in the image are available in the 'data' property.
    """

    def __init__(self, filename, path=None):
        """
        initialization of the class which will prepare the parser and parse
        the files content into class properties

        Parameters
        ----------
        filename :  str
            file name of the TIFF-like image file
        path :      str, optional
            path of the data file
        """
        if path:
            full_filename = os.path.join(path, filename)
        else:
            full_filename = filename

        with xu_open(full_filename, 'rb') as fh:
            self.byteorder = fh.read(2*dlen['char'])
            self.version = numpy.frombuffer(fh.read(dlen['word']),
                                            dtype=numpy.uint16)[0]
            if self.byteorder not in (b'II', b'MM') or self.version != 42:
                raise TypeError("Not a TIFF file (%s)" % filename)
            if self.byteorder != b'II':
                raise NotImplementedError("The 'MM' byte order is not yet "
                                          "implemented, please file a bug!")

            fh.seek(4)
            self.ifdoffset = numpy.frombuffer(fh.read(dlen['dword']),
                                              dtype=numpy.uint32)[0]
            fh.seek(self.ifdoffset)

            self.ntags = numpy.frombuffer(fh.read(dlen['word']),
                                          dtype=numpy.uint16)[0]

            self._parseImgTags(fh, self.ntags)

            fh.seek(self.ifdoffset + 2 + 12 * self.ntags)
            nextimgoffset = numpy.frombuffer(fh.read(dlen['dword']),
                                             dtype=numpy.uint32)[0]
            if nextimgoffset != 0:
                raise NotImplementedError("Multiple images per file are not "
                                          "supported, please file a bug!")

            # check if image type is supported
            if self.imgtags.get('Compression', 1) != 1:
                raise NotImplementedError("Compression is not supported, "
                                          "please file a bug report!")
            if self.imgtags.get('PhotometricInterpretation', 0) not in (0, 1):
                raise NotImplementedError("RGB and colormap is not supported")

        sf = self.imgtags.get('SampleFormat', 1)
        bs = self.imgtags.get('BitsPerSample', 1)
        if isinstance(self.imgtags['StripOffsets'], numpy.ndarray):
            hdrlen = self.imgtags['StripOffsets'][0]
        else:
            hdrlen = self.imgtags['StripOffsets']
        ImageReader.__init__(self,
                             self.imgtags['ImageLength'],
                             self.imgtags['ImageWidth'],
                             hdrlen=hdrlen,
                             dtype=tiffdtype[sf][bs],
                             byte_swap=False)

        self.data = self.readImage(filename, path)

    def _parseImgTags(self, fh, ntags):
        """
        parse TIFF image tags from Image File Directory header

        Parameters
        ----------
        fh :    file-handle
            file handle of the TIFF file
        ntags : int
            number of tags in the Image File Directory
        """

        self.imgtags = {}
        for i in range(ntags):
            ftag = numpy.frombuffer(fh.read(dlen['word']),
                                    dtype=numpy.uint16)[0]
            ftype = numpy.frombuffer(fh.read(dlen['word']),
                                     dtype=numpy.uint16)[0]
            flength = numpy.frombuffer(fh.read(dlen['dword']),
                                       dtype=numpy.uint32)[0]
            fdoffset = numpy.frombuffer(fh.read(dlen['dword']),
                                        dtype=numpy.uint32)[0]

            pos = fh.tell()
            if flength*dlen[dtypes[ftype]] <= 4:
                fdoffset = pos - dlen['dword']
            fh.seek(fdoffset)
            if ftype == 2:
                fdata = fh.read(flength * dlen[dtypes[ftype]])
                fdata = fdata.split(b'\x00')[0].decode("ascii",
                                                       errors="replace")
            else:
                rlen = flength * dlen[dtypes[ftype]]
                fdata = numpy.frombuffer(fh.read(rlen), dtype=nptyp[ftype])
            if flength == 1:
                fdata = fdata[0]
            fh.seek(pos)

            # add field to tags
            self.imgtags[tifftags.get(ftag, ftag)] = fdata


class PerkinElmer(ImageReader):
    """
    parse PerkinElmer CCD frames (``*.tif``) to numpy arrays
    Ignore the header since it seems to contain no useful data

    The routine was tested only for files with 2048x2048 pixel images
    created at Hasylab Hamburg which save an 32bit float per point.
    """

    def __init__(self, **keyargs):
        """
        initialize the PerkinElmer reader, which includes setting the dimension
        of the images as well as defining the data used for flat- and darkfield
        correction!

        Parameters
        ----------
        flatfield :     str or ndarray, optional
            filename or data for flatfield correction. supported file types
            include (.bin .bin.xz and .npy files). otherwise a 2D numpy
            array should be given
        darkfield :     str or ndarray, optional
            filename or data for darkfield correction. same types as for flat
            field are supported.
        """

        ImageReader.__init__(self, 2048, 2048, hdrlen=8, dtype=numpy.float32,
                             byte_swap=False, **keyargs)


class RoperCCD(ImageReader):

    """
    parse RoperScientific CCD frames (``*.bin``) to numpy arrays
    Ignore the header since it seems to contain no useful data

    The routine was tested only for files with 4096x4096 pixel images
    created at Hasylab Hamburg which save an 16bit integer per point.
    """

    def __init__(self, **keyargs):
        """
        initialize the RoperCCD reader, which includes setting the dimension of
        the images as well as defining the data used for flat- and darkfield
        correction!

        Parameters
        ----------
        flatfield :     str or ndarray, optional
            filename or data for flatfield correction. supported file types
            include (.bin .bin.xz and .npy files). otherwise a 2D numpy
            array should be given
        darkfield :     str or ndarray, optional
            filename or data for darkfield correction. same types as for flat
            field are supported.
        """

        ImageReader.__init__(self, 4096, 4096, hdrlen=216, dtype=numpy.int16,
                             byte_swap=False, **keyargs)


class Pilatus100K(ImageReader):
    """
    parse Dectris Pilatus 100k frames (``*.tiff``) to numpy arrays
    Ignore the header since it seems to contain no useful data
    """

    def __init__(self, **keyargs):
        """
        initialize the Piulatus100k reader, which includes setting the
        dimension of the images as well as defining the data used for flat- and
        darkfield correction!

        Parameters
        ----------
        flatfield :     str or ndarray, optional
            filename or data for flatfield correction. supported file types
            include (.bin .bin.xz and .npy files). otherwise a 2D numpy
            array should be given
        darkfield :     str or ndarray, optional
            filename or data for darkfield correction. same types as for flat
            field are supported.
        """

        ImageReader.__init__(self, 195, 487, hdrlen=4096, dtype=numpy.int32,
                             byte_swap=False, **keyargs)


def get_tiff(filename, path=None):
    """
    read tiff image file and return the data

    Parameters
    ----------
    filename :  str
        filename of the image to be read. so far only single filenames are
        supported. The data might be compressed.
    path :      str, optional
        path of the data file
    """

    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.io.get_tiff: file %s" % (filename))
        t1 = time.time()

    t = TIFFRead(filename, path=path)

    if config.VERBOSITY >= config.INFO_ALL:
        t2 = time.time()
        print("XU.io.get_tiff: parsing time %8.3f" % (t2 - t1))

    return t.data



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
from math import pi

import numpy

from .. import config, utilities
from ..math import VecNorm


def show_reciprocal_space_plane(
        mat, exp, ttmax=None, maxqout=0.01, scalef=100, ax=None, color=None,
        show_Laue=True, show_legend=True, projection='perpendicular',
        label=None):
    """
    show a plot of the coplanar diffraction plane with peak positions for the
    respective material. the size of the spots is scaled with the strength of
    the structure factor

    Parameters
    ----------
    mat:        Crystal
        instance of Crystal for structure factor calculations
    exp:        Experiment
        instance of Experiment (needs to be HXRD, or FourC for onclick action
        to work correctly). defines the inplane and out of plane direction as
        well as the sample azimuth
    ttmax:      float, optional
        maximal 2Theta angle to consider, by default 180deg
    maxqout:    float, optional
        maximal out of plane q for plotted Bragg peaks as fraction of exp.k0
    scalef:     float, or callable, optional
        scale factor or function for the marker size. If this is a function it
        should take only one float argument and return another float which is
        used as 's' parameter in matplotlib.pyplot.scatter
    ax:         matplotlib.Axes, optional
        matplotlib Axes to use for the plot, useful if multiple materials
        should be plotted in one plot
    color:      matplotlib color, optional
    show_Laue:  bool, optional
        flag to indicate if the Laue zones should be indicated
    show_legend:    bool, optional
        flag to indiate if a legend should be shown
    projection: 'perpendicular', 'polar', optional
        type of projection for Bragg peaks which do not fall into the
        diffraction plane. 'perpendicular' (default) uses only the inplane
        component in the scattering plane, whereas 'polar' uses the vectorial
        absolute value of the two inplane components. See also the 'maxqout'
        option.
    label:  None or str, optional
        label to be used for the legend. If 'None' the name of the material
        will be used.

    Returns
    -------
    Axes, plot_handle
    """
    def get_peaks(mat, exp, ttmax):
        """
        Parameters
        ----------
        mat:        Crystal
            instance of Crystal for structure factor calculations
        exp:        Experiment
            instance of Experiment (likely HXRD, or FourC)
        tt_cutoff:  float
            maximal 2Theta angle to consider, by default 180deg

        Returns
        -------
        ndarray
            data array with columns for 'q', 'qvec', 'hkl', 'r' for the Bragg
            peaks
        """
        qmax = 2 * exp.k0 * math.sin(math.radians(ttmax/2.))
        hkls = tuple(mat.lattice.get_allowed_hkl(qmax))

        q = mat.Q(hkls)
        data = numpy.zeros(len(hkls), dtype=[('qx', numpy.double),
                                             ('qy', numpy.double),
                                             ('qz', numpy.double),
                                             ('r', numpy.double),
                                             ('hkl', numpy.ndarray)])
        qvec = exp.Transform(q)
        data['qx'] = qvec[:, 0]
        data['qy'] = qvec[:, 1]
        data['qz'] = qvec[:, 2]
        rref = abs(mat.StructureFactor((0, 0, 0), exp.energy)) ** 2
        data['r'] = numpy.abs(mat.StructureFactorForQ(q, exp.energy)) ** 2
        data['r'] /= rref
        data['hkl'] = hkls

        return data

    plot, plt = utilities.import_matplotlib_pyplot('XU.materials')

    if not plot:
        print('matplotlib needed for show_reciprocal_space_plane')
        return

    if ttmax is None:
        ttmax = 180

    d = get_peaks(mat, exp, ttmax)
    k0 = exp.k0

    if not ax:
        fig = plt.figure(figsize=(9, 5))
        ax = plt.subplot(111)
    else:
        fig = ax.get_figure()
        plt.sca(ax)

    plt.axis('scaled')
    ax.set_autoscaley_on(False)
    ax.set_autoscalex_on(False)
    plt.xlim(-2.05*k0, 2.05*k0)
    plt.ylim(-0.05*k0, 2.05*k0)

    if show_Laue:
        c = plt.matplotlib.patches.Circle((0, 0), 2*k0, facecolor='#FF9180',
                                          edgecolor='none')
        ax.add_patch(c)
        qmax = 2 * k0 * math.sin(math.radians(ttmax/2.))
        c = plt.matplotlib.patches.Circle((0, 0), qmax, facecolor='#FFFFFF',
                                          edgecolor='none')
        ax.add_patch(c)

        c = plt.matplotlib.patches.Circle((0, 0), 2*k0, facecolor='none',
                                          edgecolor='0.5')
        ax.add_patch(c)
        c = plt.matplotlib.patches.Circle((k0, 0), k0, facecolor='none',
                                          edgecolor='0.5')
        ax.add_patch(c)
        c = plt.matplotlib.patches.Circle((-k0, 0), k0, facecolor='none',
                                          edgecolor='0.5')
        ax.add_patch(c)
        plt.hlines(0, -2*k0, 2*k0, color='0.5', lw=0.5)
        plt.vlines(0, -2*k0, 2*k0, color='0.5', lw=0.5)

    # mask for plotting
    m = numpy.abs(d['qx']) < maxqout*k0

    if projection == 'perpendicular':
        x = d['qy'][m]
    else:
        x = numpy.sign(d['qy'][m])*numpy.sqrt(d['qx'][m]**2 + d['qy'][m]**2)
    y = d['qz'][m]
    s = numpy.empty_like(d['r'][m])
    if callable(scalef):
        s[...] = [scalef(r) for r in d['r'][m]]
    else:
        s = d['r'][m]*scalef

    label = label if label else mat.name
    h = plt.scatter(x, y, s=s, zorder=2, label=label)
    if color:
        h.set_color(color)

    plt.xlabel(r'$Q$ inplane ($\mathrm{\AA^{-1}}$)')
    plt.ylabel(r'$Q$ out of plane ($\mathrm{\AA^{-1}}$)')

    if show_legend:
        if len(fig.legends) == 1:
            fig.legends[0].remove()
        fig.legend(*ax.get_legend_handles_labels(), loc='upper right')
    plt.tight_layout()

    annot = ax.annotate("", xy=(0, 0), xytext=(20, 20),
                        textcoords="offset points",
                        bbox=dict(boxstyle="round", fc="w"),
                        arrowprops=dict(arrowstyle="->"))
    annot.set_visible(False)

    def update_annot(ind):
        pos = h.get_offsets()[ind["ind"][0]]
        annot.xy = pos
        text = "{}\n{}".format(mat.name,
                               str(d['hkl'][m][ind['ind'][0]]))
        annot.set_text(text)
        annot.get_bbox_patch().set_facecolor(h.get_facecolor()[0])
        annot.get_bbox_patch().set_alpha(0.2)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            cont, ind = h.contains(event)
            if cont:
                update_annot(ind)
                annot.set_visible(True)
                fig.canvas.draw_idle()
            else:
                if vis:
                    annot.set_visible(False)
                    fig.canvas.draw_idle()

    def click(event):
        if event.inaxes == ax:
            cont, ind = h.contains(event)
            if cont:
                popts = numpy.get_printoptions()
                numpy.set_printoptions(precision=4, suppress=True)
                q = (d['qx'][m][ind['ind'][0]], d['qy'][m][ind['ind'][0]],
                     d['qz'][m][ind['ind'][0]])
                angles = exp.Q2Ang(q, trans=False, geometry='real')
                text = "{}\nhkl: {}\nangles: {}".format(
                    mat.name, str(d['hkl'][m][ind['ind'][0]]), str(angles))
                numpy.set_printoptions(**popts)
                print(text)

    fig.canvas.mpl_connect("motion_notify_event", hover)
    fig.canvas.mpl_connect("button_press_event", click)

    return ax, h



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2012 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2012-2020 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
script to create the HDF5 database from the raw data of XOP.
this file is only needed for administration and also used during the build

It accepts two command line arguments. The first one is the filename
of the database (required).

The second optional argument is the directory of the source data file used to
generate the database. If no path is given the 'data' sub-folder of the scripts
directory is used.
"""

import sys

# local import
from database import createAndFillDatabase


if __name__ == "__main__":
    createAndFillDatabase(*sys.argv[1:], verbose=False)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import fractions
import math
import sys

import numpy

from .. import config


def center_of_mass(pos, data, background='none', full_output=False):
    """
    function to determine the center of mass of an array

    Parameters
    ----------
    pos :  array-like
        position of the data points
    data :  array-like
        data values
    background : {'none', 'constant', 'linear'}
        type of background, either 'none', 'constant' or 'linear'
    full_output : bool
        return background cleaned data and background-parameters

    Returns
    -------
    float
        center of mass position
    """
    # subtract background
    slope = 0
    back = 0
    if background == 'linear':
        dx = float(pos[-1] - pos[0])
        if abs(dx) > 0:
            slope = (float(data[-1]) - float(data[0])) / dx
        back = (data[0] - slope * pos[0] +
                data[-1] - slope * pos[-1]) / 2.0
        ld = data - (slope * pos + back)
    elif background == 'constant':
        back = numpy.median(data)
        ld = data - numpy.min(data)
    else:
        ld = data

    ipos = numpy.sum(pos * ld) / numpy.sum(ld)
    if full_output:
        return ipos, ld, back, slope
    else:
        return ipos


def fwhm_exp(pos, data):
    """
    function to determine the full width at half maximum value of experimental
    data. Please check the obtained value visually (noise influences the
    result)

    Parameters
    ----------
    pos :   array-like
        position of the data points
    data :  array-like
        data values

    Returns
    -------
    float
        fwhm value
    """

    m = data.max()
    p0 = numpy.argmax(data)
    datal = data[:p0+1]
    datar = data[p0:]

    # determine left side half value position
    try:
        pls = pos[:p0+1][datal < m / 2.][-1]
        pll = pos[:p0+1][datal > m / 2.][0]
        ds = data[pos == pls][0]
        dl = data[pos == pll][0]
        pl = pls + (pll - pls) * (m / 2. - ds) / (dl - ds)
    except IndexError:
        if config.VERBOSITY >= config.INFO_LOW:
            print("XU.math.fwhm_exp: warning: left side half value could"
                  " not be determined -> returns 2*hwhm")
        pl = None

    # determine right side half value position
    try:
        prs = pos[p0:][datar < m / 2.][0]
        prl = pos[p0:][datar > m / 2.][-1]
        ds = data[pos == prs][0]
        dl = data[pos == prl][0]
        pr = prs + (prl - prs) * (m / 2. - ds) / (dl - ds)
    except IndexError:
        if config.VERBOSITY >= config.INFO_LOW:
            print("XU.math.fwhm_exp: warning: right side half value could"
                  " not be determined -> returns 2*hwhm")
        pr = None

    if pl is None:
        return numpy.abs(pr - p0)*2
    elif pr is None:
        return numpy.abs(pl - p0)*2
    else:
        return numpy.abs(pr - pl)


def gcd(lst):
    """
    greatest common divisor function using library functions

    Parameters
    ----------
    lst:    array-like
        array of integer values for which the greatest common divisor should be
        determined

    Returns
    -------
    gcd:    int
    """
    if numpy.version.version >= '1.15.0':
        return numpy.gcd.reduce(lst)
    elif sys.version_info >= (3, 5):
        gcdfunc = numpy.frompyfunc(math.gcd, 2, 1)
    else:
        gcdfunc = numpy.frompyfunc(fractions.gcd, 2, 1)
    return numpy.ufunc.reduce(gcdfunc, lst)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2010-2012,2014-2019
#               Dominik Kriegner <dominik.kriegner@gmail.com>
# Copyright (C) 2012 Tanja Etzelstorfer <tanja.etzelstorfer@jku.at>

# module for handling files stored in the EDF data format developed by the ESRF

import os.path
import re
import struct

import numpy

from .. import config, utilities
from .filedir import FileDirectory
from .helper import xu_h5open, xu_open

edf_kv_split = re.compile(r"\s*=\s*")  # key value sepeartor for header data
edf_eokv = re.compile(r";")  # end of line for a header
# regular expressions for several ASCII representations of numbers
edf_integer_value = re.compile(r"\d+")
edf_float_value = re.compile(r"[+-]*\d+\.*\d*")
edf_float_e_value = re.compile(r"[+-]*\d+\.\d*e[+-]*\d*")
edf_name_start_num = re.compile(r"^\d")

# dictionary mapping EDF data type keywords onto struct data types
DataTypeDict = {"SignedByte": "b",
                "SignedShort": "h",
                "SignedInteger": "i",
                "SignedLong": "i",
                "FloatValue": "f",
                "DoubleValue": "d",
                "UnsignedByte": "B",
                "UnsignedShort": "H",
                "UnsignedInt": "I",
                "UnsignedLong": "L"}
# SignedLong is only 4byte, on my 64bit machine using SignedLong:"l" caused
# troubles
# UnsignedLong is only 4byte, on my 64bit machine using UnsignedLong:"L"
# caused troubles ("I" works)


class EDFFile(object):

    def __init__(self, fname, nxkey="Dim_1", nykey="Dim_2",
                 dtkey="DataType", path="", header=True, keep_open=False):
        """
        Parameters
        ----------
        fname :	    str
            name of the EDF file of type .edf or .edf.gz

        nxkey :	    str, optional
            name of the header key that holds the number of points in
            x-direction
        nykey :	    str, optional
            name of the header key that holds the number of points in
            y-direction
        dtkey :	    str, optional
            name of the header key that holds the datatype for the binary data
        path :      str, optional
            path to the EDF file
        header :    bool, optional
            has header (default true)
        keep_open : bool, optional
            if True the file handle is kept open between multiple calls which
            can cause significant speed-ups
        """

        self.filename = fname
        self.full_filename = os.path.join(path, fname)

        # evaluate keyword arguments
        self.nxkey = nxkey
        self.nykey = nykey
        self.dtkey = dtkey
        self.headerflag = header

        # create attributes for holding data
        self._data = {}
        self._headers = []
        self._data_offsets = []
        self._data_read = False
        self._dimx = []
        self._dimy = []
        self._byte_order = []
        self._fmt_str = []
        self._dtype = []

        self.Parse()
        if keep_open:
            self.fid = xu_open(self.full_filename, 'rb')
        else:
            self.fid = None

        self.nimages = len(self._data_offsets)
        self.header = self._headers[0]

    def Parse(self):
        """
        Parse file to find the number of entries and read the respective
        header information
        """
        header = {}
        offset = 0

        with xu_open(self.full_filename, 'rb') as fid:
            if config.VERBOSITY >= config.INFO_ALL:
                print("XU.io.EDFFile.Parse: file: %s" % self.full_filename)

            if self.headerflag:
                while True:  # until end of file
                    hdr_flag = False
                    ml_value_flag = False  # marks a multiline header
                    for line in fid:  # until end of header
                        linelength = len(line)
                        offset += linelength
                        line = line.decode('ascii', 'ignore')
                        if config.VERBOSITY >= config.DEBUG:
                            print(line)
                        if line == "":
                            break
                        # remove leading and trailing whitespace symbols
                        line = line.strip()

                        if line == "{" and not hdr_flag:
                            # start with header
                            hdr_flag = True
                            header = {}
                            continue

                        if hdr_flag:
                            # stop reading when the end of the header
                            # is reached
                            if line == "}":
                                # place offset reading here - here we get the
                                # real starting position of the binary data!!
                                break

                            # continue if the line has no content
                            if line == "":
                                continue

                            # split key and value of the header entry
                            if not ml_value_flag:
                                try:
                                    key, value = edf_kv_split.split(line, 1)
                                except ValueError:
                                    print("XU.io.EDFFile.Parse: "
                                          "line: %s" % line)

                                key = key.strip()
                                value = value.strip()

                                # if the value extends over multiple lines set
                                # the multiline value flag
                                if value[-1] != ";":
                                    ml_value_flag = True
                                else:
                                    value = value[:-1]
                                    value = value.strip()
                                    header[key] = value
                            else:
                                value = value + line
                                if value[-1] == ";":
                                    ml_value_flag = False

                                    value = value[:-1]
                                    value = value.strip()
                                    header[key] = value
                    else:
                        break
                    # append header to class variables
                    self._byte_order.append(header["ByteOrder"])
                    self._fmt_str.append(DataTypeDict[header[self.dtkey]])
                    self._dimx.append(int(header[self.nxkey]))
                    self._dimy.append(int(header[self.nykey]))
                    self._dtype.append(header[self.dtkey])

                    self._headers.append(header)
                    self._data_offsets.append(offset)
                    # jump over data block
                    tot_nofp = self._dimx[-1] * self._dimy[-1]
                    dsize = tot_nofp * struct.calcsize(self._fmt_str[-1])
                    fid.seek(offset + dsize, 0)
                    offset += dsize

            else:  # in case of no header also save one set of defaults
                self._byte_order.append('LowByteFirst')
                self._fmt_str.append(DataTypeDict['UnsignedShort'])
                self._dimx.append(516)
                self._dimy.append(516)
                self._dtype.append('UnsignedShort')
                self._headers.append(header)
                self._data_offsets.append(offset)

        # try to parse motor positions and counters from last found header
        # into separate dictionary
        if 'motor_mne' in header:
            tkeys = header['motor_mne'].split()
            try:
                tval = numpy.array(header['motor_pos'].split(),
                                   dtype=numpy.double)
                self.motors = dict(zip(tkeys, tval))
            except ValueError:
                print("XU.io.EDFFile.ReadData: Warning: header conversion "
                      "of motor positions failed")

        if 'counter_mne' in header:
            tkeys = header['counter_mne'].split()
            try:
                tval = numpy.array(header['counter_pos'].split(),
                                   dtype=numpy.double)
                self.counters = dict(zip(tkeys, tval))
            except ValueError:
                print("XU.io.EDFFile.ReadData: Warning: header conversion "
                      "of counter values failed")

    def ReadData(self, nimg=0):
        """
        Read the CCD data of the specified image and return the data
        this function is called automatically when the 'data' property is
        accessed, but can also be called manually when only a certain image
        from the file is needed.

        Parameters
        ----------
        nimg :      int, optional
            number of the image which should be read (starts with 0)
        """
        if self.fid:
            binfid = self.fid
            # move to the data section - jump over the header
            binfid.seek(self._data_offsets[nimg], 0)
            # read the data
            tot_nofp = self._dimx[nimg] * self._dimy[nimg]
            fmt_str = self._fmt_str[nimg]
            bindata = binfid.read(tot_nofp * struct.calcsize(fmt_str))
        else:
            with xu_open(self.full_filename, 'rb') as binfid:
                # move to the data section - jump over the header
                binfid.seek(self._data_offsets[nimg], 0)
                # read the data
                tot_nofp = self._dimx[nimg] * self._dimy[nimg]
                fmt_str = self._fmt_str[nimg]
                bindata = binfid.read(tot_nofp * struct.calcsize(fmt_str))
        if config.VERBOSITY >= config.DEBUG:
            print("XU.io.EDFFile: read binary data: nofp: %d len: %d"
                  % (tot_nofp, len(bindata)))
            print("XU.io.EDFFile: format: %s" % fmt_str)

        try:
            data = numpy.frombuffer(bindata, count=tot_nofp, dtype=fmt_str)
        except ValueError:
            if fmt_str == 'L':
                fmt_str = 'I'
                try:
                    data = numpy.frombuffer(bindata, count=tot_nofp,
                                            dtype=fmt_str)
                except ValueError:
                    raise IOError("XU.io.EDFFile: data format (%s) has "
                                  "different byte-length, from amount of data "
                                  "one expects %d bytes per entry"
                                  % (fmt_str, len(bindata) / tot_nofp))
            else:
                raise IOError("XU.io.EDFFile: data format (%s) has different "
                              "byte-length, from amount of data one expects "
                              "%d bytes per entry"
                              % (fmt_str, len(bindata) / tot_nofp))

        data.shape = (self._dimy[nimg], self._dimx[nimg])

        if self._byte_order[nimg] != "LowByteFirst":  # data = data.byteswap()
            print("XU.io.EDFFile.ReadData: check byte order - "
                  "not low byte first")

        return data

    @property
    def data(self):
        if not self._data_read:
            for i in range(self.nimages):
                self._data[i] = self.ReadData(i)
            self._data_read = True
        if self.nimages == 1:
            return self._data[0]
        else:
            return self._data

    def Save2HDF5(self, h5f, group="/", comp=True):
        """
        Saves the data stored in the EDF file in a HDF5 file as a HDF5 array.
        By default the data is stored in the root group of the HDF5 file - this
        can be changed by passing the name of a target group or a path to the
        target group via the "group" keyword argument.

        Parameters
        ----------
        h5f :	    file-handle or str
            a HDF5 file object or name
        group :     str, optional
            group where to store the data (default to the root of the file)
        comp :	    bool, optional
            activate compression - true by default
        """
        with xu_h5open(h5f, 'a') as h5:
            if isinstance(group, str):
                if group == '/':
                    g = h5
                else:
                    if group in h5:
                        del h5[group]
                    g = h5.create_group(group)
            else:
                g = group

            # create the array name
            ca_name = os.path.split(self.filename)[-1]
            ca_name = os.path.splitext(ca_name)[0]
            # perform a second time for case of .edf.gz files
            ca_name = os.path.splitext(ca_name)[0]
            ca_name = utilities.makeNaturalName(ca_name)
            if edf_name_start_num.match(ca_name):
                ca_name = "ccd_" + ca_name
            if config.VERBOSITY >= config.INFO_ALL:
                print(ca_name)

            # create the array description
            ca_desc = "EDF CCD data from file %s " % (self.filename)
            kwds = {'fletcher32': True}
            if comp:
                kwds['compression'] = 'gzip'

            if self.nimages != 1:
                ca_name += '_{n:04d}'

            for n in range(self.nimages):
                d = self.ReadData(n)
                name = ca_name.format(n=n)
                try:
                    ca = g.create_dataset(name, data=d, **kwds)
                except ValueError:
                    del g[name]
                    ca = g.create_dataset(name, data=d, **kwds)

                ca.attrs['TITLE'] = ca_desc

                # finally we have to append the attributes
                for k in self.header:
                    ca.attrs[utilities.makeNaturalName(k)] = self.header[k]


class EDFDirectory(FileDirectory):

    """
    Parses a directory for EDF files, which can be stored to a HDF5 file for
    further usage
    """

    def __init__(self, datapath, ext="edf", **keyargs):
        """

        Parameters
        ----------
        datapath :	str
            directory of the EDF file
        ext :           str, optional
            extension of the ccd files in the datapath (default: "edf")
        keyargs :       dict, optional
            further keyword arguments are passed to EDFFile
        """
        super().__init__(datapath, ext, EDFFile, **keyargs)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
Panalytical XML (www.XRDML.com) data file parser

based on the native python xml.dom.minidom module.
want to keep the number of dependancies as small as possible
"""

import os.path
import warnings
from xml.etree import cElementTree as ElementTree

import numpy

from .. import config
from .helper import xu_open


class XRDMLMeasurement(object):

    """
    class to handle scans in a XRDML datafile
    """

    def __init__(self, measurement, namespace=''):
        """
        initialization routine for a XRDML measurement which parses are all
        scans within this measurement.
        """

        self.namespace = namespace
        # get scans in <xrdMeasurement>
        slist = measurement.findall(self.namespace + "scan")

        self.hkl = (numpy.nan, numpy.nan, numpy.nan)
        self.material = ""
        self.ddict = {}
        for field in ["countTime", "detector", "counts",
                      "beamAttenuationFactors", "hkl"]:
            self.ddict[field] = []
        is_scalar = 0

        # loop over all scan entries - scan points
        for s in slist:
            # check if scan is complete
            scanstatus = s.get("status")
            if scanstatus in ("Aborted", "Not finished") and len(slist) > 1:
                if config.VERBOSITY >= config.INFO_LOW:
                    print("XU.io.XRDMLFile: subscan has been aborted "
                          "(part of the data unavailable)!")
            else:
                self.scanmotname = s.get("scanAxis")
                reflection = s.find(self.namespace + "reflection")
                if reflection:
                    m = reflection.find(self.namespace + "material")
                    if m is not None:
                        self.material = m.text
                    hkl = reflection.find(self.namespace + "hkl")
                    if hkl:
                        hkl_h = int(hkl.find(self.namespace + "h").text)
                        hkl_k = int(hkl.find(self.namespace + "k").text)
                        hkl_l = int(hkl.find(self.namespace + "l").text)
                        self.hkl = (hkl_h, hkl_k, hkl_l)
                points = s.find(self.namespace + "dataPoints")

                # add count time to output data
                countTime = points.find(self.namespace +
                                        "commonCountingTime").text
                self.ddict["countTime"].append(float(countTime))

                # check for intensities first to get number of points in scan
                int_elem = points.find(self.namespace + "intensities")
                if int_elem is not None:
                    data = int_elem.text
                    hascounts = False
                else:
                    ct_elem = points.find(self.namespace + "counts")
                    ct_npy = numpy.fromstring(ct_elem.text, sep=" ")
                    self.ddict["counts"].append(ct_npy.tolist())
                    data = ct_elem.text
                    hascounts = True
                # count time normalization; output is counts/sec
                ct_rate = numpy.fromstring(data, sep=" ") / float(countTime)
                nofpoints = ct_rate.size
                # if present read beamAttenuationFactors
                # they are already corrected in the data file, but may be
                # interesting
                attfact = points.find(self.namespace +
                                      "beamAttenuationFactors")
                if attfact is not None:
                    atten = numpy.fromstring(attfact.text, sep=" ")
                    atten_list = atten.tolist()
                    self.ddict["beamAttenuationFactors"].append(atten_list)
                    hasatten = True
                else:
                    hasatten = False

                if hascounts and hasatten:
                    self.ddict["detector"].append((ct_rate*atten).tolist())
                else:
                    self.ddict["detector"].append(ct_rate.tolist())

                # read the axes position
                pos = points.findall(self.namespace + "positions")
                for p in pos:
                    # read axis name and unit
                    aname = p.get("axis")
                    # aunit = p.get("unit")

                    # read axis data
                    listp = p.findall(self.namespace + "listPositions")
                    s = p.findall(self.namespace + "startPosition")
                    e = p.findall(self.namespace + "endPosition")
                    c = p.find(self.namespace + "commonPosition")
                    if listp:  # listPositions
                        listp = listp[0]
                        data_list = numpy.fromstring(listp.text, sep=" ")
                        data_list = data_list.tolist()
                    elif s and e:  # start endPosition
                        data_list = numpy.linspace(
                            float(s[0].text), float(e[0].text),
                            nofpoints).tolist()
                    elif c is not None:  # commonPosition
                        data_list = numpy.fromstring(c.text, sep=" ")
                        data_list = data_list.tolist()
                        is_scalar = 1
                    else:
                        raise ValueError(
                            "no positions for axis {} found".format(aname))

                    # have to append the data to the data dictionary in case
                    # the scan is complete!
                    if aname not in self.ddict:
                        self.ddict[aname] = []
                    if not is_scalar:
                        self.ddict[aname].append(data_list)
                    else:
                        self.ddict[aname].append(data_list[0])
                        is_scalar = 0

        # finally all scan data needs to be converted to numpy arrays
        for k in self.ddict:
            self.ddict[k] = numpy.array(self.ddict[k])

        # flatten output if only one scan was present
        if len(slist) == 1:
            for k in self.ddict:
                self.ddict[k] = numpy.ravel(self.ddict[k])

        # save scanmot-values and detector counts in special arrays
        if self.scanmotname in ['2Theta-Omega', 'Gonio']:
            self.scanmot = self.ddict['2Theta']
        elif self.scanmotname == 'Omega-2Theta':
            self.scanmot = self.ddict['Omega']
        elif self.scanmotname in self.ddict:
            self.scanmot = self.ddict[self.scanmotname]
        else:
            warnings.warn('XU.io: unknown scan motor name in XRDML-File')
        self.int = self.ddict['detector']

    def __getitem__(self, key):
        return self.ddict[key]

    def __str__(self):
        ostr = "XRDML Measurement\n"
        if self.material:
            ostr += "Material: '%s'; hkl: %s\n" % (self.material,
                                                   str(self.hkl))
        for k in self.ddict:
            ostr += "%s with %s points\n" % (k, str(self.ddict[k].shape))

        return ostr


class XRDMLFile(object):

    """
    class to handle XRDML data files. The class is supplied with a file
    name and uses the XRDMLScan class to parse the xrdMeasurement in the
    file
    """

    def __init__(self, fname, path=""):
        """
        initialization routine supplied with a filename
        the file is automatically parsed and the data are available
        in the "scan" object. If more <xrdMeasurement> tags are present, which
        should not be the case, their data is present in the "scans" object.

        Parameters
        ----------
        fname :     str
            filename of the XRDML file
        path :      str, optional
            path to the XRDML file
        """
        self.full_filename = os.path.join(path, fname)
        self.filename = os.path.basename(self.full_filename)
        with xu_open(self.full_filename) as fid:
            d = ElementTree.parse(fid)
        root = d.getroot()
        try:
            namespace = root.tag[:root.tag.index('}')+1]
        except ValueError:
            namespace = ''

        slist = root.findall(namespace+"xrdMeasurement")

        # determine the number of scans in the file
        self.nscans = len(slist)
        self.scans = []
        for s in slist:
            self.scans.append(XRDMLMeasurement(s, namespace))

        if self.nscans == 1:
            self.scan = self.scans[0]

    def __str__(self):
        ostr = "XRDML File: %s\n" % self.filename
        for s in self.scans:
            ostr += s.__str__()

        return ostr


def getxrdml_map(filetemplate, scannrs=None, path=".", roi=None):
    """
    parses multiple XRDML file and concatenates the results for parsing the
    xrayutilities.io.XRDMLFile class is used. The function can be used for
    parsing maps measured with the PIXCel 1D detector (and in limited way also
    for data acquired with a point detector -> see getxrdml_scan instead).

    Parameters
    ----------
    filetemplate :  str
        template string for the file names, can contain a %d which is replaced
        by the scan number or be a list of filenames
    scannrs :       int or list, optional
        scan number(s)
    path :          str, optional
        common path to the filenames
    roi :           tuple, optional
        region of interest for the PIXCel detector, for other measurements this
        is not useful!

    Returns
    -------
    om, tt, psd :   ndarray
        motor positions and data as flattened numpy arrays

    Examples
    --------
    >>> om, tt, psd = xrayutilities.io.getxrdml_map("samplename_%d.xrdml",
    >>>                                             [1, 2], path="./data")
    """
    def getOmPixcel(omraw, ttraw):
        """
        function to reshape the Omega values into a form needed for
        further treatment with xrayutilities
        """
        return (omraw[:, numpy.newaxis] * numpy.ones(ttraw.shape)).flatten()

    # read raw data and convert to reciprocal space
    om = numpy.zeros(0)
    tt = numpy.zeros(0)
    psd = numpy.zeros(0)
    # create scan names
    if scannrs is None:
        files = [filetemplate]
    else:
        files = list()
        if not getattr(scannrs, '__iter__', False):
            scannrs = [scannrs]
        for nr in scannrs:
            files.append(filetemplate % nr)

    # parse files
    for f in files:
        d = XRDMLFile(os.path.join(path, f))
        s = d.scan
        if len(s['detector'].shape) == 1:
            raise TypeError("XU.getxrdml_map: This function can only be used "
                            "to parse reciprocal space map files")

        if roi is None:
            roi = [0, s['detector'].shape[1]]
        if s['Omega'].size < s['2Theta'].size:
            om = numpy.concatenate(
                (om, getOmPixcel(s['Omega'], s['2Theta'][:, roi[0]:roi[1]])))
            tt = numpy.concatenate(
                (tt, s['2Theta'][:, roi[0]:roi[1]].flatten()))
        elif s['Omega'].size > s['2Theta'].size:
            om = numpy.concatenate((om, s['Omega'].flatten()))
            tt = numpy.concatenate((
                tt,
                numpy.ravel(s['2Theta'][:, numpy.newaxis] *
                            numpy.ones(s['Omega'].shape))))
        else:
            om = numpy.concatenate((om, s['Omega'].flatten()))
            tt = numpy.concatenate(
                (tt, s['2Theta'][:, roi[0]:roi[1]].flatten()))
        psd = numpy.concatenate(
            (psd, s['detector'][:, roi[0]:roi[1]].flatten()))

    return om, tt, psd


def getxrdml_scan(filetemplate, *motors, **kwargs):
    """
    parses multiple XRDML file and concatenates the results for parsing the
    xrayutilities.io.XRDMLFile class is used. The function can be used for
    parsing arbitrary scans and will return the the motor values of the scan
    motor and additionally the positions of the motors given by in the
    ``*motors`` argument

    Parameters
    ----------
    filetemplate :  str
        template string for the file names, can contain a %d which is replaced
        by the scan number or be a list of filenames given by the scannrs
        keyword argument

    motors :        str
        motor names to return: e.g.: 'Omega', '2Theta', ...  one can also use
        abbreviations:

         - 'Omega' = 'om' = 'o'
         - '2Theta' = 'tt' = 't'
         - 'Chi' = 'c'
         - 'Phi' = 'p'

    scannrs :       int or list, optional
        scan number(s)
    path :          str, optional
        common path to the filenames

    Returns
    -------
    scanmot, mot1, mot2,..., detectorint :  ndarray
        motor positions and data as flattened numpy arrays

    Examples
    --------
    >>> scanmot, om, tt, inte = xrayutilities.io.getxrdml_scan(
    >>>     "samplename_1.xrdml", 'om', 'tt', path="./data")
    """
    flatten = True
    # parse keyword arguments
    path = kwargs.get('path', '.')
    scannrs = kwargs.get('scannrs', None)

    validmotors = ['Omega', '2Theta', 'Psi', 'Chi', 'Phi', 'Z', 'X', 'Y']
    validmotorslow = [mot.lower() for mot in validmotors]
    # create correct motor names from input values
    motnames = []
    for mot in motors:
        if mot.lower() in validmotorslow:
            motnames.append(validmotors[validmotorslow.index(mot.lower())])
        elif mot.lower() in ['phi', 'p']:
            motnames.append('Phi')
        elif mot.lower() in ['chi', 'c']:
            motnames.append('Chi')
        elif mot.lower() in ['psi']:
            motnames.append('Psi')
        elif mot.lower() in ['tt', 't']:
            motnames.append('2Theta')
        elif mot.lower() in ['om', 'o']:
            motnames.append('Omega')
        else:
            raise ValueError("XU: invalid motor name given")

    motvals = numpy.empty((len(motnames) + 1, 0))
    detvals = numpy.empty(0)
    # create scan names
    if scannrs is None:
        if isinstance(filetemplate, list):
            files = filetemplate
        else:
            files = [filetemplate]
    else:
        files = list()
        if not numpy.iterable(scannrs):
            scannrs = [scannrs]
        for nr in scannrs:
            files.append(filetemplate % nr)

    # parse files
    if len(files) == 1:
        flatten = False
    for f in files:
        d = XRDMLFile(os.path.join(path, f))
        s = d.scan
        detshape = s['detector'].shape

        if len(detshape) == 2:
            angles = numpy.ravel(s.scanmot)
            angles.shape = (1, angles.size)
            for mot in motnames:
                if s[mot].shape != detshape:
                    angles = numpy.vstack((
                        angles,
                        numpy.ravel(s[mot][:, numpy.newaxis] *
                                    numpy.ones(detshape))))
                else:
                    angles = numpy.vstack((angles, numpy.ravel(s[mot])))
            motvals = numpy.concatenate((motvals, angles), axis=1)
            dval = numpy.ravel(s['detector'])
            detvals = numpy.concatenate((detvals, dval))
            if not flatten:
                detvals.shape = detshape
                motvals.shape = (len(motnames) + 1, detshape[0], detshape[1])
        else:
            detvals = numpy.concatenate((detvals, s['detector']))
            angles = s.scanmot
            angles.shape = (1, angles.size)
            for mot in motnames:
                try:
                    angles = numpy.vstack((angles, s[mot]))
                except ValueError:  # motor is not array
                    angles = numpy.vstack(
                        (angles, s[mot] * numpy.ones(detshape)))
            motvals = numpy.concatenate((motvals, angles), axis=1)

    # make return value
    ret = []
    for i in range(motvals.shape[0]):
        ret.append(motvals[i, ...])
    ret.append(detvals)
    return ret



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015-2019 Dominik Kriegner <dominik.kriegner@gmail.com>


"""
class for reading data + header information from Rigaku RAS (3-column ASCII)
files

Such datafiles are generated by the Smartlab Guidance software from Rigaku.
"""

import os.path
import re
from itertools import islice

import numpy
import numpy.lib.recfunctions

from .. import config
from ..exception import InputError
# relative imports from xrayutilities
from .helper import xu_open

re_measstart = re.compile(r"^\*RAS_DATA_START")
re_measend = re.compile(r"^\*RAS_DATA_END")
re_headerstart = re.compile(r"^\*RAS_HEADER_START")
re_headerend = re.compile(r"^\*RAS_HEADER_END")
re_datastart = re.compile(r"^\*RAS_INT_START")
re_dataend = re.compile(r"^\*RAS_INT_END")
re_scanaxis = re.compile(r"^\*MEAS_SCAN_AXIS_X_INTERNAL")
re_intstart = re.compile(r"^\*RAS_INT_START")
re_datestart = re.compile(r"^\*MEAS_SCAN_START_TIME")
re_datestop = re.compile(r"^\*MEAS_SCAN_END_TIME")
re_initmoponame = re.compile(r"^\*MEAS_COND_AXIS_NAME_INTERNAL")
re_initmopovalue = re.compile(r"^\*MEAS_COND_AXIS_POSITION")
re_datacount = re.compile(r"^\*MEAS_DATA_COUNT")
re_measspeed = re.compile(r"^\*MEAS_SCAN_SPEED ")
re_measstep = re.compile(r"^\*MEAS_SCAN_STEP ")


class RASFile(object):

    """
    Represents a RAS data file. The file is read during the
    constructor call

    Parameters
    ----------
    filename :  str
        name of the ras-file
    path :      str, optional
        path to the data file
    """

    def __init__(self, filename, path=None):
        self.filename = filename
        if path is None:
            self.full_filename = self.filename
        else:
            self.full_filename = os.path.join(path, self.filename)

        self.scans = []
        self.Read()

    def Read(self):
        """
        Read the data from the file
        """
        with xu_open(self.full_filename) as fid:
            while True:
                t = fid.tell()
                line = fid.readline()
                line = line.decode('ascii', 'ignore')
                if config.VERBOSITY >= config.DEBUG:
                    print("XU.io.RASFile: %d: '%s'" % (t, line))
                if re_measstart.match(line):
                    continue
                elif re_headerstart.match(line):
                    s = RASScan(self.full_filename, t)
                    self.scans.append(s)
                    fid.seek(s.fidend)  # set handle to after scan
                elif re_measend.match(line) or line in (None, ''):
                    break
                else:
                    continue
        if len(self.scans) > 0:
            self.scan = self.scans[0]


class RASScan(object):

    """
    Represents a single Scan portion of a RAS data file. The scan is parsed
    during the constructor call

    Parameters
    ----------
    filename :  str
        file name of the data file
    pos :       int
        seek position of the 'RAS_HEADER_START' line
    """

    def __init__(self, filename, pos):
        self.filename = filename
        self.fidpos = pos
        self.fidend = pos
        with xu_open(self.filename) as self.fid:
            self.fid.seek(self.fidpos)
            self._parse_header()
            self._parse_data()
            self.fidend = self.fid.tell()

    def _parse_header(self):
        """
        Read the data from the file
        """
        # read header
        self.header = []
        keys = {}
        position = {}
        offset = self.fid.tell()
        for line in self.fid:
            offset += len(line)
            line = line.decode('ascii', 'ignore')
            self.header.append(line)
            if config.VERBOSITY >= config.DEBUG:
                print("XU.io.RASScan: %d: '%s'" % (offset, line))

            if re_datestart.match(line):
                m = line.split(' ', 1)[-1].strip()
                self.scan_start = m.strip('"')
            elif re_datestop.match(line):
                m = line.split(' ', 1)[-1].strip()
                self.scan_stop = m.strip('"')
            elif re_initmoponame.match(line):
                idx = int(line.split('-', 1)[-1].split()[0])
                moname = line.split(' ', 1)[-1].strip().strip('"')
                keys[idx] = moname
            elif re_initmopovalue.match(line):
                idx = int(line.split('-', 1)[-1].split()[0])
                mopos = line.split(' ', 1)[-1].strip().strip('"')
                try:
                    mopos = float(mopos)
                except ValueError:
                    pass
                position[idx] = mopos
            elif re_scanaxis.match(line):
                self.scan_axis = line.split(' ', 1)[-1].strip().strip('"')
            elif re_datacount.match(line):
                length = line.split(' ', 1)[-1].strip().strip('"')
                self.length = int(float(length))
            elif re_measspeed.match(line):
                speed = line.split(' ', 1)[-1].strip().strip('"')
                self.meas_speed = float(speed)
            elif re_measstep.match(line):
                step = line.split(' ', 1)[-1].strip().strip('"')
                self.meas_step = float(step)
            elif re_headerend.match(line):
                break

        # generate header dictionary
        self.init_mopo = {}
        for k in keys:
            self.init_mopo[keys[k]] = position[k]
        self.fid.seek(offset)

    def _parse_data(self):
        line = self.fid.readline().decode('ascii', 'ignore')
        offset = self.fid.tell()
        if re_datastart.match(line):
            lines = islice(self.fid, self.length)
            self.data = numpy.genfromtxt(lines)
            self.data = numpy.rec.fromrecords(self.data,
                                              names=[self.scan_axis,
                                                     'int',
                                                     'att'])
            self.fid.seek(offset)
            lines = islice(self.fid, self.length)
            dlength = numpy.sum([len(line) for line in lines])
            if config.VERBOSITY >= config.DEBUG:
                print("XU.io.RASScan: offset %d; data-length %d"
                      % (offset, dlength))
            self.fid.seek(offset + dlength)
        else:
            raise IOError('File handle at wrong position to read data!')


def getras_scan(scanname, scannumbers, *args, **kwargs):
    """
    function to obtain the angular cooridinates as well as intensity values
    saved in RAS datafiles. Especially useful for reciprocal space map
    measurements, and to combine date from several scans

    further more it is possible to obtain even more positions from
    the data file if more than two string arguments with its names are given

    Parameters
    ----------
    scanname :      str
        name of the scans, for multiple scans this needs to be a template
        string
    scannumbers :   int, tuple or list
        number of the scans of the reciprocal space map
    args :          str, optional
        names of the motors. to read reciprocal space maps measured in coplanar
        diffraction give:

            - omname:  name of the omega motor (or its equivalent)
            - ttname:  name of the two theta motor (or its equivalent)

    kwargs :        dict
        keyword arguments forwarded to RASFile function

    Returns
    -------
    [ang1, ang2, ...] :     list
        angular positions are extracted from the respective scan header, or
        motor positions during the scan. this is omitted if no `args` are given
    rasdata :   ndarray
        the data values (includes the intensities e.g. rasdata['int']).

    Examples
    --------
    >>> [om, tt], MAP = xu.io.getras_scan('text%05d.ras', 36, 'Omega',
    >>>                                   'TwoTheta')
    """

    if isinstance(scannumbers, (list, tuple)):
        scanlist = scannumbers
    else:
        scanlist = list([scannumbers])

    angles = dict.fromkeys(args)
    for key in angles:
        if not isinstance(key, str):
            raise InputError("*arg values need to be strings with motornames")
        angles[key] = numpy.zeros(0)
    buf = numpy.zeros(0)
    MAP = numpy.zeros(0)

    for nr in scanlist:
        rasfile = RASFile(scanname % nr, **kwargs)
        for scan in rasfile.scans:
            sdata = scan.data
            if MAP.dtype == numpy.float64:
                MAP.dtype = sdata.dtype
            # append scan data to MAP, where all data are stored
            MAP = numpy.append(MAP, sdata)
            # check type of scan
            for i in range(len(args)):
                motname = args[i]
                scanlength = len(sdata)
                try:
                    buf = sdata[motname]
                except ValueError:
                    buf = scan.init_mopo[motname] * numpy.ones(scanlength)
                angles[motname] = numpy.concatenate((angles[motname], buf))

    retval = []
    for motname in args:
        # create return values in correct order
        retval.append(angles[motname])

    if not args:
        return MAP
    elif len(args) == 1:
        return retval[0], MAP
    else:
        return retval, MAP



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2021 Dominik Kriegner <dominik.kriegner@gmail.com>
import abc
import collections.abc
import copy
import warnings

import numpy
from scipy.constants import physical_constants
from scipy.misc import derivative

from .. import materials, utilities
from ..math import heaviside
from .models import LayerModel


def getit(it, key):
    """
    generator to obtain items from nested iterable
    """
    for elem in it:
        if isinstance(elem, collections.abc.Iterable):
            if key in elem:
                yield elem[key]
            else:
                for subelem in getit(elem, key):
                    yield subelem


def getfirst(iterable, key):
    """
    helper function to obtain the first item in a nested iterable
    """
    return next(getit(iterable, key))


def GradedBuffer(xfrom, xto, nsteps, thickness, relaxation=1):
    """
    create a multistep graded composition buffer.

    Parameters
    ----------
    xfrom :     float
        begin of the composition gradient
    xto :       float
        end of the composition gradient
    nsteps :    int
        number of steps of the gradient
    thickness : float
        total thickness of the Buffer in A
    relaxation : float
        relaxation of the buffer

    Returns
    -------
    list
        layer list needed for the Darwin model simulation
    """
    subthickness = thickness/nsteps
    gradedx = numpy.linspace(xfrom, xto, nsteps)
    layerlist = []
    for x in gradedx:
        layerlist.append({'t': subthickness, 'x': x, 'r': relaxation})
    return layerlist


class DarwinModel(LayerModel):
    """
    model class inmplementing the basics of the Darwin theory for layers
    materials.  This class is not fully functional and should be used to derive
    working models for particular material systems.

    To make the class functional the user needs to implement the
    init_structurefactors() and _calc_mono() methods
    """
    ncalls = 0

    def __init__(self, qz, qx=0, qy=0, **kwargs):
        """
        constructor of the model class. The arguments consist of basic
        parameters which are needed to prepare the calculation of the model.

        Parameters
        ----------
        qz :        array-like
            momentum transfer values for the calculation
        qx, qy :    float, optional
            inplane momentum transfer (not implemented!)
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        resolution_width :  float, optional
            width of the resolution function (deg)
        polarization : {'S', 'P', 'both'}
            polarization of the x-ray beam. If set to 'both' also Cmono, the
            polarization factor of the monochromator should be set
        experiment : Experiment, optional
            experiment class containing geometry and energy of the experiment.
        Cmono :     float, optional
            polarization factor of the monochromator
        energy :    float or str, optional
            x-ray energy in eV
        """
        self.polarization = kwargs.pop('polarization', 'S')
        exp = kwargs.pop('experiment', None)
        self.Cmono = kwargs.pop('Cmono', 1)
        super().__init__(exp, **kwargs)

        self.npoints = len(qz)
        self.qz = numpy.asarray(qz)
        self.qinp = (qx, qy)
        if self.qinp != (0, 0):
            raise NotImplementedError('asymmetric CTR simulation is not yet '
                                      'supported -> approach the authors')
        self.init_structurefactors()
        # initialize coplanar geometry
        k = self.exp.k0
        qv = numpy.asarray([(qx, qy, q) for q in self.qz])
        Q = numpy.linalg.norm(qv, axis=1)
        theta = numpy.arcsin(Q / (2 * k))
        domega = numpy.arctan2(numpy.sqrt(qx**2 + qy**2), self.qz)
        self.alphai, self.alphaf = (theta + domega, theta - domega)
        # polarization factor
        self.C = {'S': numpy.ones(len(self.qz)),
                  'P': numpy.abs(numpy.cos(self.alphai + self.alphaf))}

    def init_structurefactors(self):
        """
        calculates the needed atomic structure factors
        """
        pass

    def _calc_mono(self, pdict, pol):
        """
        calculate the reflection and transmission coefficients of monolayer

        Parameters
        ----------
        pdict :     dict
            property dictionary, contains all the properties for the structure
            factor calculation
        pol :       {'S', 'P'}
            polarization of the x-rays; sigma or pi

        Returns
        -------
        r, rbar, t :    float or array-like
            reflection, backside reflection, and tranmission coefficients
        """
        pass

    def _calc_double(self, ra, rabar, ta, rb, rbbar, tb, d):
        """
        calculate reflection coefficients for the double layer from the
        reflection values of the single layers

        Parameters
        ----------
        ra, rabar, ta : float or array-like
            reflection, backside reflection, and transmission coefficients of
            layer A
        rb, rbbar, tb : float or array-like
            same for layer B
        d :             float
            distance between the layers

        Returns
        -------
        r, rbar, t :    float or array-like
            reflection, backside reflection, and tranmission coefficients
        """
        self.ncalls += 1
        e = numpy.exp(-1j*self.qz*d)
        eh = numpy.exp(-1j*self.qz*d/2)
        denom = (1-rabar*rb*e)
        rab = ra + rb*(ta*ta*e)/denom
        rabbar = rbbar + rabar*(tb*tb*e)/(1-rbbar*ra*e)
        tab = ta*tb*eh/denom
        return rab, rabbar, tab

    def simulate(self, ml):
        """
        main simulation function for the Darwin model. will calculate the
        reflected intensity

        Parameters
        ----------
        ml :        iterable
            monolayer sequence of the sample. This should be created with the
            function make_monolayer(). see its documentation for details
        """
        self.ncalls = 0
        Ih = {'S': numpy.zeros(len(self.qz)), 'P': numpy.zeros(len(self.qz))}
        geomfact = heaviside(self.alphai) * heaviside(self.alphaf)
        for pol in self.get_polarizations():
            r, rbar, t = (numpy.zeros(self.npoints, dtype=complex),
                          numpy.zeros(self.npoints, dtype=complex),
                          numpy.ones(self.npoints, dtype=complex))
            for nrep, subml in ml:
                r, rbar, t = self._recur_sim(nrep, subml, r, rbar, t, pol)
            self.r, self.rbar, self.t = r, rbar, t
            Ih[pol] = numpy.abs(self.r)**2 * geomfact

        ret = self.join_polarizations(Ih['S'], Ih['P'])
        return self._create_return(self.qz, numpy.sqrt(ret))

    def _recur_sim(self, nrep, ml, r, rbar, t, pol):
        """
        recursive simulation function for the calculation of the reflected,
        backside reflected and transmitted wave factors (internal)

        Parameters
        ----------
        ml :        iterable
            monolayer sequence of the calculation block. This should be created
            with the function make_monolayer(). see its documentation for
            details
        r :         float or array-like
            reflection factor of the upper layers (needed for the recursion)
        rbar :      float or array-like
            back-side reflection factor of the upper layers (needed for the
            recursion)
        t :         float or array-like
            transmission factor of the upper layers (needed for the recursion)
        pol :       {'S', 'P'}
            polarization of the x-rays

        Returns
        -------
        r, rbar, t : float or array-like
            reflection and transmission of the full stack
        """
        if isinstance(ml, list):
            rm, rmbar, tm = (None, None, None)
            for nsub, subml in ml:
                rm, rmbar, tm = self._recur_sim(nsub, subml, rm,
                                                rmbar, tm, pol)
            d = getfirst(ml, 'd')
        else:
            rm, rmbar, tm = self._calc_mono(ml, pol)
            d = ml['d']

        Nmax = int(numpy.log(nrep) / numpy.log(2)) + 1
        for i in range(Nmax):
            if r is None:
                r, rbar, t = rm, rmbar, tm
            elif (nrep // (2**i)) % 2 == 1:
                r, rbar, t = self._calc_double(r, rbar, t, rm, rmbar, tm, d)
            rm, rmbar, tm = self._calc_double(rm, rmbar, tm, rm, rmbar, tm, d)

        return r, rbar, t


class DarwinModelAlloy(DarwinModel, utilities.ABC):
    """
    extension of the DarwinModel for an binary alloy system were one parameter
    is used to determine the chemical composition

    To make the class functional the user needs to implement the
    get_dperp_apar() method and define the substrate lattice parameter (asub).
    See the DarwinModelSiGe001 class for an implementation example.
    """
    @abc.abstractmethod
    def get_dperp_apar(self, x, apar, r=1):
        """
        calculate inplane lattice parameter and the out of plane lattice plane
        spacing (of the atomic planes!) from composition and relaxation.

        Parameters
        ----------
        x :         float
            chemical composition parameter
        apar :      float
            inplane lattice parameter of the material below the current layer
            (onto which the present layer is strained to). This value also
            served as a reference for the relaxation parameter.
        r :         float
            relaxation parameter. 1=relaxed, 0=pseudomorphic

        Returns
        -------
        dperp :     float
        apar :      float
        """
        raise NotImplementedError("abstract method needs to be overwritten")

    def make_monolayers(self, s):
        """
        create monolayer sequence from layer list

        Parameters
        ----------
        s :     list
            layer model. list of layer dictionaries including possibility to
            form superlattices. As an example 5 repetitions of a
            Si(10nm)/Ge(15nm) superlattice on Si would like like:

            >>> s = [(5, [{'t': 100, 'x': 0, 'r': 0},
            >>>           {'t': 150, 'x': 1, 'r': 0}]),
            >>>      {'t': 3500000, 'x': 0, 'r': 0}]

            the dictionaries must contain 't': thickness in A, 'x': chemical
            composition, and either 'r': relaxation or 'ai': inplane lattice
            parameter.
            Future implementations for asymmetric peaks might include layer
            type 'l' (not yet implemented). Already now any additional property
            in the dictionary will be handed on to the returned monolayer list.
        asub :  float
            inplane lattice parameter of the substrate

        Returns
        -------
        list
            monolayer list in a format understood by the simulate and
            xGe_profile methods
        """
        ml = []
        ai = self.asub
        for subl in copy.copy(s):
            ml, ai = self._recur_makeml(subl, ml, ai)
        return ml

    def _recur_makeml(self, s, ml, apar):
        """
        recursive creation of a monolayer structure (internal)

        Parameters
        ----------
        s :     list
            layer model. list of layer dictionaries
        ml :    list
            list of layers below what should be added (needed for recursion)
        apar :  float
            inplane lattice parameter of the current surface

        Returns
        -------
        list
            monolayer list in a format understood by the simulate and
            prop_profile methods
        """

        if isinstance(s, tuple):
            nrep, sd = s
            if isinstance(sd, dict):
                sd = [sd, ]
            if any([r > 0 for r in getit(sd, 'r')]):  # if relaxation
                for i in range(nrep):
                    for subsd in sd:
                        ml, apar = self._recur_makeml(subsd, ml, apar=apar)
            else:  # no relaxation in substructure
                subl = []
                for subsd in sd:
                    subl, apar = self._recur_makeml(subsd, subl, apar=apar)
                ml.insert(0, (nrep, subl))
        elif isinstance(s, dict):
            x = s.pop('x')
            if callable(x):  # composition profile in layer
                t = 0
                T = s.pop('t')
                if 'r' in s:
                    if s['r'] > 0:
                        warnings.warn(
                            """relaxation for composition gradient may yield
                            weird lattice parameter variation! Consider
                            supplying the inplane lattice parameter 'ai'
                            directly!""")
                while t < T:
                    if 'r' in s:
                        r = abs(derivative(x, t, dx=1.4, n=1))*s['r']
                        dperp, apar = self.get_dperp_apar(x(t), apar, r)
                    else:
                        dperp, apar = self.get_dperp_apar(x(t), s['ai'])
                    t += dperp
                    d = copy.copy(s)
                    d.pop('r')
                    d.update({'d': dperp, 'x': x(t), 'ai': apar})
                    ml.insert(0, (1, d))
            else:  # constant composition layer
                if 'r' in s:
                    dperp, apar = self.get_dperp_apar(x, apar, s.pop('r'))
                else:
                    dperp, apar = self.get_dperp_apar(x, s.pop('ai'))
                nmono = int(numpy.ceil(s['t']/dperp))
                s.update({'d': dperp, 'x': x, 'ai': apar})
                ml.insert(0, (nmono, s))
        else:
            raise Exception('wrong type (%s) of sublayer, must be tuple or'
                            ' dict' % (type(s)))
        return ml, apar

    def prop_profile(self, ml, prop):
        """
        calculate the profile of chemical composition or inplane lattice
        spacing from a monolayer list. One value for each monolayer in the
        sample is returned.

        Parameters
        ----------
        ml :    list
            monolayer list created by make_monolayer()
        prop :  str
            name of the property which should be evaluated. Use 'x' for the
            chemical composition and 'ai' for the inplane lattice parameter.

        Returns
        -------
        zm :    ndarray
            z-position, z-0 is the surface
        propx : ndarray
            value of the property prop for every monolayer
        """

        def startinterval(start, inter, N):
            return numpy.arange(start, start+inter*(N+0.5), inter)

        def _recur_prop(nrep, ml, zp, propx, propn):
            if isinstance(ml, list):
                lzp, lprop = ([], [])
                for nreps, subml in ml:
                    lzp, lprop = _recur_prop(nreps, subml, lzp, lprop, propn)
            else:
                lzp = -ml['d']
                lprop = ml[propn]

            Nmax = int(numpy.log(nrep) / numpy.log(2)) + 1
            for i in range(Nmax):
                if (nrep // (2**i)) % 2 == 1:
                    try:
                        curzp = zp[-1]
                    except IndexError:
                        curzp = 0.0
                    zp = numpy.append(zp, lzp+curzp)
                    propx = numpy.append(propx, lprop)
                try:
                    curlzp = lzp[-1]
                except IndexError:
                    curlzp = lzp
                lzp = numpy.append(lzp, lzp+curlzp)
                lprop = numpy.append(lprop, lprop)
            return zp, propx

        zm = []
        propx = []
        for nrep, subml in ml:
            zm, propx = _recur_prop(nrep, subml, zm, propx, prop)
        return zm, propx


class DarwinModelSiGe001(DarwinModelAlloy):
    """
    model class implementing the Darwin theory of diffraction for SiGe layers.
    The model is based on separation of the sample structure into building
    blocks of atomic planes from which a multibeam dynamical model is
    calculated.
    """
    Si = materials.Si
    Ge = materials.Ge
    eSi = materials.elements.Si
    eGe = materials.elements.Ge
    aSi = materials.Si.a1[0]
    asub = aSi  # needed for the make_monolayer function
    re = physical_constants['classical electron radius'][0] * 1e10

    @classmethod
    def abulk(cls, x):
        """
        calculate the bulk (relaxed) lattice parameter of the alloy
        """
        return cls.aSi + (0.2 * x + 0.027 * x ** 2)

    @staticmethod
    def poisson_ratio(x):
        """
        calculate the Poisson ratio of the alloy
        """
        return 2 * (63.9-15.6*x) / (165.8-37.3*x)  # according to IOFFE

    @classmethod
    def get_dperp_apar(cls, x, apar, r=1):
        """
        calculate inplane lattice parameter and the out of plane lattice plane
        spacing (of the atomic planes!) from composition and relaxation

        Parameters
        ----------
        x :     float
            chemical composition parameter
        apar :  float
            inplane lattice parameter of the material below the current layer
            (onto which the present layer is strained to). This value also
            served as a reference for the relaxation parameter.
        r :     float, optional
            relaxation parameter. 1=relaxed, 0=pseudomorphic

        Returns
        -------
        dperp : float
            perpendicular d-spacing
        apar :  float
            inplane lattice parameter
        """
        abulk = cls.abulk(x)
        aparl = apar + (abulk - apar) * r
        dperp = abulk*(1+cls.poisson_ratio(x)*(1-aparl/abulk))/4.
        return dperp, aparl

    def init_structurefactors(self, temp=300):
        """
        calculates the needed atomic structure factors

        Parameters
        ----------
        temp :  float, optional
            temperature used for the Debye model
        """
        en = self.exp.energy
        q = numpy.sqrt(self.qinp[0]**2 + self.qinp[1]**2 + self.qz**2)
        self.fSi = self.eSi.f(q, en) * self.Si._debyewallerfactor(temp, q)
        self.fGe = self.eGe.f(q, en) * self.Ge._debyewallerfactor(temp, q)
        self.fSi0 = self.eSi.f(0, en)
        self.fGe0 = self.eGe.f(0, en)

    def _calc_mono(self, pdict, pol):
        """
        calculate the reflection and transmission coefficients of monolayer

        Parameters
        ----------
        pdict :     dict
            property dictionary, contains the layer properties:
            'x':   Ge-content of the layer (0: Si, 1: Ge);
            'l':   index of the layer in the unit cell (0, 1, 2, 3). important
                   for asymmetric peaks only!
        pol :       {'S', 'P'}
            polarization of the x-rays

        Returns
        -------
        r, rbar, t : float or array-like
            reflection, backside reflection, and tranmission coefficients
        """
        ainp = pdict.get('ai')
        xGe = pdict.get('x')
        # pre-factor for reflection: contains footprint correction
        gamma = 4*numpy.pi*self.re/(self.qz*ainp**2)
#        ltype = pdict.get('l', 0)
#        if ltype == 0: # for asymmetric peaks (not yet implemented)
#            p1, p2 = (0, 0), (0.5, 0.5)
#        elif ltype == 1:
#            p1, p2 = (0.25, 0.25), (0.75, 0.75)
#        elif ltype == 2:
#            p1, p2 = (0.5, 0.), (0., 0.5)
#        elif ltype == 3:
#            p1, p2 = (0.75, 0.25), (0.25, 0.75)

        r = -1j*gamma * self.C[pol] * (self.fSi+(self.fGe-self.fSi)*xGe) * 2
        # * (exp(1j*(h*p1[0]+k*p1[1])) + exp(1j*(h*p1[0]+k*p1[1])))
        t = 1 + 1j*gamma * (self.fSi0+(self.fGe0-self.fSi0)*xGe) * 2
        return r, numpy.copy(r), t


class DarwinModelGaInAs001(DarwinModelAlloy):
    """
    Darwin theory of diffraction for Ga_{1-x} In_x As layers.
    The model is based on separation of the sample structure into building
    blocks of atomic planes from which a multibeam dynamical model is
    calculated.
    """
    GaAs = materials.GaAs
    InAs = materials.InAs
    eGa = materials.elements.Ga
    eIn = materials.elements.In
    eAs = materials.elements.As
    aGaAs = materials.GaAs.a1[0]
    asub = aGaAs  # needed for the make_monolayer function
    re = physical_constants['classical electron radius'][0] * 1e10

    @classmethod
    def abulk(cls, x):
        """
        calculate the bulk (relaxed) lattice parameter of the Ga_{1-x}In_{x}As
        alloy
        """
        return cls.aGaAs + 0.40505*x

    @staticmethod
    def poisson_ratio(x):
        """
        calculate the Poisson ratio of the alloy
        """
        return 2 * (4.54 + 0.8*x) / (8.34 + 3.56*x)  # according to IOFFE

    @classmethod
    def get_dperp_apar(cls, x, apar, r=1):
        """
        calculate inplane lattice parameter and the out of plane lattice plane
        spacing (of the atomic planes!) from composition and relaxation

        Parameters
        ----------
        x :     float
            chemical composition parameter
        apar :  float
            inplane lattice parameter of the material below the current layer
            (onto which the present layer is strained to). This value also
            served as a reference for the relaxation parameter.
        r :     float
            relaxation parameter. 1=relaxed, 0=pseudomorphic

        Returns
        -------
        dperp : float
            perpendicular d-spacing
        apar :  float
            inplane lattice parameter
        """
        abulk = cls.abulk(x)
        aparl = apar + (abulk - apar) * r
        dperp = abulk*(1+cls.poisson_ratio(x)*(1-aparl/abulk))/4.
        return dperp, aparl

    def init_structurefactors(self, temp=300):
        """
        calculates the needed atomic structure factors

        Parameters
        ----------
        temp :      float, optional
            temperature used for the Debye model
        """
        en = self.exp.energy
        q = numpy.sqrt(self.qinp[0]**2 + self.qinp[1]**2 + self.qz**2)
        fAs = self.eAs.f(q, en)
        self.fGaAs = (self.eGa.f(q, en) + fAs) \
            * self.GaAs._debyewallerfactor(temp, q)
        self.fInAs = (self.eIn.f(q, en) + fAs) \
            * self.InAs._debyewallerfactor(temp, q)
        self.fGaAs0 = self.eGa.f(0, en) + self.eAs.f(0, en)
        self.fInAs0 = self.eIn.f(0, en) + self.eAs.f(0, en)

    def _calc_mono(self, pdict, pol):
        """
        calculate the reflection and transmission coefficients of monolayer

        Parameters
        ----------
        pdict :     dict
            property dictionary, contains the layer properties:
            'x':   In-content of the layer (0: GaAs, 1: InAs)
        pol :       {'S', 'P'}
            polarization of the x-rays

        Returns
        -------
        r, rbar, t : float or array-like
            reflection, backside reflection, and tranmission coefficients
        """
        ainp = pdict.get('ai')
        xInAs = pdict.get('x')
        # pre-factor for reflection: contains footprint correction
        gamma = 4*numpy.pi * self.re/(self.qz*ainp**2)
        r = -1j*gamma*self.C[pol]*(self.fGaAs+(self.fInAs-self.fGaAs)*xInAs)
        t = 1 + 1j*gamma * (self.fGaAs0+(self.fInAs0-self.fGaAs0)*xInAs)
        return r, numpy.copy(r), t


class DarwinModelAlGaAs001(DarwinModelAlloy):
    """
    Darwin theory of diffraction for Al_x Ga_{1-x} As layers.
    The model is based on separation of the sample structure into building
    blocks of atomic planes from which a multibeam dynamical model is
    calculated.
    """
    GaAs = materials.GaAs
    AlAs = materials.AlAs
    eGa = materials.elements.Ga
    eAl = materials.elements.Al
    eAs = materials.elements.As
    aGaAs = materials.GaAs.a1[0]
    asub = aGaAs  # needed for the make_monolayer function
    re = physical_constants['classical electron radius'][0] * 1e10

    @classmethod
    def abulk(cls, x):
        """
        calculate the bulk (relaxed) lattice parameter of the Al_{x}Ga_{1-x}As
        alloy
        """
        return cls.aGaAs + 0.0078*x

    @staticmethod
    def poisson_ratio(x):
        """
        calculate the Poisson ratio of the alloy
        """
        return 2 * (5.38+0.32*x) / (11.88+0.14*x)  # according to IOFFE

    @classmethod
    def get_dperp_apar(cls, x, apar, r=1):
        """
        calculate inplane lattice parameter and the out of plane lattice plane
        spacing (of the atomic planes!) from composition and relaxation

        Parameters
        ----------
        x :     float
            chemical composition parameter
        apar :  float
            inplane lattice parameter of the material below the current layer
            (onto which the present layer is strained to). This value also
            served as a reference for the relaxation parameter.
        r :     float
            relaxation parameter. 1=relaxed, 0=pseudomorphic

        Returns
        -------
        dperp : float
            perpendicular d-spacing
        apar :  float
            inplane lattice parameter
        """
        abulk = cls.abulk(x)
        aparl = apar + (abulk - apar) * r
        dperp = abulk*(1+cls.poisson_ratio(x)*(1-aparl/abulk))/4.
        return dperp, aparl

    def init_structurefactors(self, temp=300):
        """
        calculates the needed atomic structure factors

        Parameters
        ----------
        temp :  float, optional
            temperature used for the Debye model
        """
        en = self.exp.energy
        q = numpy.sqrt(self.qinp[0]**2 + self.qinp[1]**2 + self.qz**2)
        fAs = self.eAs.f(q, en)
        self.fGaAs = (self.eGa.f(q, en) + fAs) \
            * self.GaAs._debyewallerfactor(temp, q)
        self.fAlAs = (self.eAl.f(q, en) + fAs) \
            * self.AlAs._debyewallerfactor(temp, q)
        self.fGaAs0 = self.eGa.f(0, en) + self.eAs.f(0, en)
        self.fAlAs0 = self.eAl.f(0, en) + self.eAs.f(0, en)

    def _calc_mono(self, pdict, pol):
        """
        calculate the reflection and transmission coefficients of monolayer

        Parameters
        ----------
        pdict :     dict
            property dictionary, contains the layer properties:
            'x':   Al-content of the layer (0: GaAs, 1: AlAs)
        pol :       {'S', 'P'}
            polarization of the x-rays

        Returns
        -------
        r, rbar, t : float or array-like
            reflection, backside reflection, and tranmission coefficients
        """
        ainp = pdict.get('ai')
        xAlAs = pdict.get('x')
        # pre-factor for reflection: contains footprint correction
        gamma = 4*numpy.pi * self.re/(self.qz*ainp**2)
        r = -1j*gamma*self.C[pol]*(self.fGaAs+(self.fAlAs-self.fGaAs)*xAlAs)
        t = 1 + 1j*gamma * (self.fGaAs0+(self.fAlAs0-self.fGaAs0)*xAlAs)
        return r, numpy.copy(r), t
